from __future__ import annotations

import logging
import os
import re
import warnings
from typing import Any

import libsbml
import pandas as pd

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", message="pkg_resources is deprecated")
    from fs import open_fs
from pydantic import RootModel, field_validator

from napistu import consensus, identifiers, sbml_dfs_utils, source, utils
from napistu.constants import (
    BQB,
    ONTOLOGIES,
    SBML_DFS,
    SBML_DFS_SCHEMA,
    SCHEMA_DEFS,
)
from napistu.ingestion.constants import (
    COMPARTMENT_ALIASES,
    COMPARTMENTS_GO_TERMS,
    GENERIC_COMPARTMENT,
    SBML_DEFS,
    VALID_COMPARTMENTS,
)

logger = logging.getLogger(__name__)


class SBML:
    """A class for handling Systems Biology Markup Language (SBML) files.

    This class provides an interface to read and parse SBML files, offering
    methods to access the model, summarize its contents, and report any errors
    encountered during parsing.

    Parameters
    ----------
    sbml_path : str
        The file path to an SBML model. Supports local paths and GCS URIs.

    Attributes
    ----------
    document : libsbml.SBMLDocument
        The raw SBML document object from libsbml.
    model : libsbml.Model
        The parsed SBML model object from libsbml.
    verbose : bool, default=False
        If True, then include detailed logs.

    Methods
    -------
    summary()
        Prints a summary of the sbml model
    sbml_errors(reduced_log, return_df)
        Print a summary of all errors in the SBML file

    Raises
    ------
    ValueError
        If the SBML model is not Level 3, or if critical, unknown errors are
        found during parsing.
    """

    def __init__(
        self,
        sbml_path: str,
        verbose: bool = False,
    ) -> None:
        """Initializes the SBML object by reading and validating an SBML file."""
        reader = libsbml.SBMLReader()
        if os.path.exists(sbml_path):
            self.document = reader.readSBML(sbml_path)
        else:
            with open_fs(os.path.dirname(sbml_path)) as fs:
                txt = fs.readtext(os.path.basename(sbml_path))
                self.document = reader.readSBMLFromString(txt)

        if self.document.getLevel() < 3:
            raise ValueError(
                f"SBML model is level {self.document.getLevel()}, only SBML 3 is supported"
            )

        self.model = self.document.getModel()

        # check for critical sbml errors
        errors = self.sbml_errors(reduced_log=False, return_df=True)
        if errors is not None:
            critical_errors = errors[errors[SBML_DEFS.ERROR_SEVERITY] >= 2]
            critical_errors = set(critical_errors[SBML_DEFS.ERROR_DESCRIPTION].unique())
            known_errors = {
                "<layout> must have 'id' and may have 'name'",
                "Missing value for the 'compartment' attribute",
            }

            found_known_errors = known_errors.intersection(critical_errors)
            if len(found_known_errors) > 0:
                if verbose:
                    logger.info(
                        f"The following known errors were found: {found_known_errors}"
                    )

            unknown_critical_errors = critical_errors - known_errors
            if len(unknown_critical_errors) != 0:
                critical_errors = ", ".join(unknown_critical_errors)
                raise ValueError(
                    f"Critical errors were found when reading the sbml file: {critical_errors}"
                )

    def sbml_errors(self, reduced_log: bool = True, return_df: bool = False):
        """Formats and reports all errors found in the SBML file.

        Parameters
        ----------
        reduced_log : bool, optional
            If True, aggregates errors by category and severity. Defaults to True.
        return_df : bool, optional
            If True, returns a DataFrame of the errors. Otherwise, prints a
            styled summary. Defaults to False.

        Returns
        -------
        pd.DataFrame or None
            A DataFrame containing the error log if `return_df` is True and
            errors are present, otherwise None.
        """
        n_errors = self.document.getNumErrors()
        if n_errors == 0:
            return None

        error_log = list()
        for i in range(n_errors):
            e = self.document.getError(i)

            error_entry = {
                SBML_DEFS.ERROR_NUMBER: i,
                SBML_DEFS.ERROR_CATEGORY: e.getCategoryAsString(),
                SBML_DEFS.ERROR_SEVERITY: e.getSeverity(),
                SBML_DEFS.ERROR_DESCRIPTION: e.getShortMessage(),
                SBML_DEFS.ERROR_MESSAGE: e.getMessage(),
            }

            error_log.append(error_entry)
        error_log = pd.DataFrame(error_log)

        if reduced_log:
            error_log = (
                error_log[
                    [
                        SBML_DEFS.ERROR_CATEGORY,
                        SBML_DEFS.ERROR_SEVERITY,
                        SBML_DEFS.ERROR_MESSAGE,
                    ]
                ]
                .groupby([SBML_DEFS.ERROR_CATEGORY, SBML_DEFS.ERROR_SEVERITY])
                .count()
            )

        if return_df:
            return error_log
        else:
            if reduced_log:
                headers = [
                    f"{SBML_DEFS.ERROR_CATEGORY}, {SBML_DEFS.ERROR_SEVERITY}",
                    "count",
                ]
            else:
                headers = [
                    SBML_DEFS.ERROR_CATEGORY,
                    SBML_DEFS.ERROR_SEVERITY,
                    SBML_DEFS.ERROR_DESCRIPTION,
                ]
                error_log = error_log[headers]

            utils.show(error_log, headers=headers)

            return None

    def summary(self) -> pd.DataFrame:
        """Generates a styled summary of the SBML model.

        Returns
        -------
        pd.io.formats.style.Styler
            A styled pandas DataFrame containing a summary of the model,
            including pathway name, ID, and counts of species and reactions.
        """
        model = self.model

        model_summaries = dict()

        model_summaries[SBML_DEFS.SUMMARY_PATHWAY_NAME] = model.getName()
        model_summaries[SBML_DEFS.SUMMARY_PATHWAY_ID] = model.getId()

        model_summaries[SBML_DEFS.SUMMARY_N_SPECIES] = model.getNumSpecies()
        model_summaries[SBML_DEFS.SUMMARY_N_REACTIONS] = model.getNumReactions()

        compartments = [
            model.getCompartment(i).getName() for i in range(model.getNumCompartments())
        ]
        compartments.sort()
        model_summaries[SBML_DEFS.SUMMARY_COMPARTMENTS] = ",\n".join(compartments)

        model_summaries_dat = pd.DataFrame(model_summaries, index=[0]).T

        return utils.show(model_summaries_dat)  # type: ignore

    def _define_compartments(
        self, compartment_aliases_dict: dict | None = None
    ) -> pd.DataFrame:
        """Extracts and defines compartments from the SBML model.

        This function iterates through the compartments in the SBML model,
        extracting their IDs, names, and identifiers. It also handles cases where
        CVTerms are missing by mapping compartment names to known GO terms.

        Parameters
        ----------
        sbml_model : SBML
            The SBML model to process.
        compartment_aliases_dict : dict, optional
            A dictionary to map custom compartment names. If None, the default
            mapping from `COMPARTMENT_ALIASES` is used.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing information about each compartment, indexed by
            compartment ID.
        """
        if compartment_aliases_dict is None:
            aliases = COMPARTMENT_ALIASES
        else:
            aliases = CompartmentAliasesValidator.from_dict(compartment_aliases_dict)

        compartments = list()
        for i in range(self.model.getNumCompartments()):
            comp = self.model.getCompartment(i)

            if not comp.getCVTerms():
                logger.warning(
                    f"Compartment {comp.getId()} has empty CVterms, mapping its c_Identifiers from the Compartment dict"
                )

                compartments.append(_define_compartments_missing_cvterms(comp, aliases))

            else:
                compartments.append(
                    {
                        SBML_DFS.C_ID: comp.getId(),
                        SBML_DFS.C_NAME: comp.getName(),
                        SBML_DFS.C_IDENTIFIERS: identifiers.cv_to_Identifiers(comp),
                        SBML_DFS.C_SOURCE: source.Source.empty(),
                    }
                )

        if len(compartments) == 0:
            logger.warning(
                "No compartments were found in the SBML model. using default compartment GO CELLULAR_COMPONENT"
            )
            return sbml_dfs_utils.stub_compartments(with_source=True)

        return pd.DataFrame(compartments).set_index(SBML_DFS.C_ID)

    def _define_cspecies(self) -> pd.DataFrame:
        """Creates a DataFrame of compartmentalized species from an SBML model.

        This function extracts all species from the model and creates a
        standardized DataFrame that includes unique IDs for each compartmentalized
        species (`sc_id`), along with species and compartment IDs, and their
        corresponding identifiers.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing information about each compartmentalized species.
        """
        comp_species = list()
        for i in range(self.model.getNumSpecies()):
            spec = self.model.getSpecies(i)

            spec_dict = {
                SBML_DFS.SC_ID: spec.getId(),
                SBML_DFS.SC_NAME: spec.getName(),
                SBML_DFS.C_ID: spec.getCompartment(),
                SBML_DFS.S_IDENTIFIERS: identifiers.cv_to_Identifiers(spec),
                SBML_DFS.SC_SOURCE: source.Source.empty(),
            }

            comp_species.append(spec_dict)

        # add geneproducts defined using L3 FBC extension
        fbc_gene_products = self._define_fbc_gene_products()
        comp_species.extend(fbc_gene_products)

        comp_species_df = pd.DataFrame(comp_species).set_index(SBML_DFS.SC_ID)
        comp_species_df[SBML_DFS.SC_NAME] = utils.update_pathological_names(
            comp_species_df[SBML_DFS.SC_NAME], "SC"
        )

        # replace empty strings for c_id with None so the resolve function will properly fill these
        comp_species_df[SBML_DFS.C_ID] = comp_species_df[SBML_DFS.C_ID].replace(
            "", None
        )

        # if all entries are missing then c_id was likely stubbed with sbml_dfs_utils.stub_compartments()
        if all(comp_species_df[SBML_DFS.C_ID].isnull()):
            default_cid = sbml_dfs_utils.stub_compartments().index[0]
            comp_species_df[SBML_DFS.C_ID] = default_cid

        return comp_species_df

    def _define_fbc_gene_products(self) -> list[dict]:

        mplugin = self.model.getPlugin("fbc")

        fbc_gene_products = list()
        if mplugin is not None:
            for i in range(mplugin.getNumGeneProducts()):
                gene_product = mplugin.getGeneProduct(i)

                gene_dict = {
                    SBML_DFS.SC_ID: gene_product.getId(),
                    SBML_DFS.SC_NAME: (
                        gene_product.getName()
                        if gene_product.isSetName()
                        else gene_product.getLabel()
                    ),
                    # use getLabel() to accomendate sbml model (e.g. HumanGEM.xml) with no fbc:name attribute
                    # Recon3D.xml has both fbc:label and fbc:name attributes, with gene name in fbc:nam
                    SBML_DFS.C_ID: None,
                    SBML_DFS.S_IDENTIFIERS: identifiers.cv_to_Identifiers(gene_product),
                    SBML_DFS.SC_SOURCE: source.Source.empty(),
                }

                fbc_gene_products.append(gene_dict)

        return fbc_gene_products

    def _define_reactions(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Extracts and defines reactions and their participating species.

        This function iterates through all reactions in the SBML model, creating
        a DataFrame for reaction attributes and another for all participating
        species (reactants, products, and modifiers).

        Parameters
        ----------
        sbml_model : SBML
            The SBML model to process.

        Returns
        -------
        tuple[pd.DataFrame, pd.DataFrame]
            A tuple containing two DataFrames:
            - The first DataFrame contains reaction attributes, indexed by reaction ID.
            - The second DataFrame lists all species participating in reactions.
        """
        reactions_list = []
        reaction_species_list = []
        for i in range(self.model.getNumReactions()):
            rxn = SBML_reaction(self.model.getReaction(i))
            reactions_list.append(rxn.reaction_dict)

            rxn_specs = rxn.species
            rxn_specs[SBML_DFS.R_ID] = rxn.reaction_dict[SBML_DFS.R_ID]
            reaction_species_list.append(rxn_specs)

        reactions = pd.DataFrame(reactions_list).set_index(SBML_DFS.R_ID)

        reaction_species_df = pd.concat(reaction_species_list)
        # add an index if reaction species didn't have IDs in the .sbml
        if all([v == "" for v in reaction_species_df.index.tolist()]):
            reaction_species_df = (
                reaction_species_df.reset_index(drop=True)
                .assign(
                    rsc_id=sbml_dfs_utils.id_formatter(
                        range(reaction_species_df.shape[0]), SBML_DFS.RSC_ID
                    )
                )
                .set_index(SBML_DFS.RSC_ID)
            )

        return reactions, reaction_species_df

    def _define_species(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Extracts and defines species and compartmentalized species.

        This function creates two DataFrames: one for unique molecular species
        (un-compartmentalized) and another for compartmentalized species, which
        represent a species within a specific compartment.

        Returns
        -------
        tuple[pd.DataFrame, pd.DataFrame]
            A tuple containing two DataFrames:
            - The first DataFrame represents unique molecular species.
            - The second DataFrame represents compartmentalized species.
        """

        SPECIES_SCHEMA = SBML_DFS_SCHEMA.SCHEMA[SBML_DFS.SPECIES]
        CSPECIES_SCHEMA = SBML_DFS_SCHEMA.SCHEMA[SBML_DFS.COMPARTMENTALIZED_SPECIES]
        SPECIES_VARS = SPECIES_SCHEMA[SCHEMA_DEFS.VARS]
        CSPECIES_VARS = CSPECIES_SCHEMA[SCHEMA_DEFS.VARS]

        comp_species_df = self._define_cspecies()

        # find unique species and create a table
        consensus_species_df = comp_species_df.copy()
        consensus_species_df.index.names = [SBML_DFS.S_ID]
        consensus_species, species_lookup = consensus._reduce_to_consensus_ids(
            consensus_species_df,
            # note that this is an incomplete schema because consensus_species_df isn't a
            # normal species table
            {
                SCHEMA_DEFS.PK: SBML_DFS.S_ID,
                SCHEMA_DEFS.ID: SBML_DFS.S_IDENTIFIERS,
                SCHEMA_DEFS.TABLE: SBML_DFS.SPECIES,
            },
        )

        # create a table of unique molecular species
        consensus_species.index.name = SBML_DFS.S_ID
        consensus_species[SBML_DFS.S_NAME] = [
            re.sub("\\[.+\\]", "", x).strip()
            for x in consensus_species[SBML_DFS.SC_NAME]
        ]
        consensus_species = consensus_species.drop(
            [SBML_DFS.SC_NAME, SBML_DFS.C_ID], axis=1
        )
        consensus_species[SBML_DFS.S_SOURCE] = [
            source.Source.empty() for x in range(0, consensus_species.shape[0])
        ]

        species = consensus_species[SPECIES_VARS]
        compartmentalized_species = comp_species_df.join(species_lookup).rename(
            columns={"new_id": SBML_DFS.S_ID}
        )[CSPECIES_VARS]

        # Internal validation: Check consistency between species and compartmentalized species
        _validate_species_consistency(species, compartmentalized_species)

        return species, compartmentalized_species


class CompartmentAliasesValidator(RootModel):
    """
    A Pydantic model for validating compartment alias dictionaries.

    This model ensures that the compartment alias dictionary is a mapping
    from a string (the canonical compartment name) to a list of strings
    (the aliases for that compartment). It also validates that the keys
    of the dictionary are valid compartment names.

    Attributes
    ----------
    root : dict[str, list[str]]
        The root of the model is a dictionary where keys are strings and
        values are lists of strings.
    """

    root: dict[str, list[str]]

    @field_validator("root")
    def validate_aliases(cls, values: dict[str, list[str]]):
        """Validate the compartment alias dictionary."""
        for key, alias_list in values.items():
            if not key:
                raise ValueError("Compartment keys must be non-empty.")
            if key not in VALID_COMPARTMENTS:
                raise ValueError(
                    f"Invalid compartment key: {key}. "
                    f"Must be one of {VALID_COMPARTMENTS}"
                )
            if not alias_list:
                raise ValueError(f"Alias list for '{key}' cannot be empty.")
        return values

    @classmethod
    def from_dict(cls, data: dict[str, list[str]]) -> "CompartmentAliasesValidator":
        """
        Create a CompartmentAliasesValidator from a dictionary.

        Parameters
        ----------
        data : dict[str, list[str]]
            A dictionary mapping canonical compartment names to their aliases.

        Returns
        -------
        CompartmentAliasesValidator
            A validated instance of the model.
        """
        return cls.model_validate(data)

    def __getitem__(self, key: str) -> list[str]:
        return self.root[key]

    def items(self):
        return self.root.items()

    def __iter__(self):
        return iter(self.root)

    def __len__(self):
        return len(self.root)


class SBML_reaction:
    """A convenience class for processing individual SBML reactions.

    This class extracts and organizes key information about an SBML reaction,
    including its attributes and participating species (substrates, products,
    and modifiers).

    Parameters
    ----------
    sbml_reaction : libsbml.Reaction
        A libsbml Reaction object to be processed.

    Attributes
    ----------
    reaction_dict : dict
        A dictionary of reaction-level attributes, including its ID, name,
        reversibility, identifiers, and source information.
    species : pd.DataFrame
        A DataFrame listing all species participating in the reaction,
        including their roles (substrate, product, modifier), stoichiometry,
        and SBO terms.
    """

    def __init__(
        self,
        sbml_reaction: libsbml.Reaction,
    ) -> None:
        """Initializes the SBML_reaction object by parsing a libsbml Reaction."""
        reaction_dict = {
            SBML_DFS.R_ID: sbml_reaction.getId(),
            SBML_DFS.R_NAME: sbml_reaction.getName(),
            SBML_DFS.R_IDENTIFIERS: identifiers.cv_to_Identifiers(sbml_reaction),
            SBML_DFS.R_SOURCE: source.Source.empty(),
            SBML_DFS.R_ISREVERSIBLE: sbml_reaction.getReversible(),
        }

        self.reaction_dict = reaction_dict

        # process reaction species
        reaction_species = list()
        # save modifiers
        for i in range(sbml_reaction.getNumModifiers()):
            spec = sbml_reaction.getModifier(i)
            spec_dict = {
                SBML_DFS.RSC_ID: spec.getId(),
                SBML_DFS.SC_ID: spec.getSpecies(),
                SBML_DFS.STOICHIOMETRY: 0,
                SBML_DFS.SBO_TERM: spec.getSBOTermID(),
            }
            reaction_species.append(spec_dict)

        # find gene products defined using the fbc plugin
        rxn_fbc = sbml_reaction.getPlugin("fbc")
        if rxn_fbc:
            gpa = rxn_fbc.getGeneProductAssociation()
            if gpa:
                gene_products = _extract_gene_products(gpa.getAssociation())
                # de-duplicate
                gene_products = list(
                    {d[SBML_DFS.SC_ID]: d for d in gene_products}.values()
                )
                reaction_species.extend(gene_products)

        # save reactants
        for i in range(sbml_reaction.getNumReactants()):
            spec = sbml_reaction.getReactant(i)
            spec_dict = {
                SBML_DFS.RSC_ID: spec.getId(),
                SBML_DFS.SC_ID: spec.getSpecies(),
                SBML_DFS.STOICHIOMETRY: -1 * spec.getStoichiometry(),
                SBML_DFS.SBO_TERM: spec.getSBOTermID(),
            }
            reaction_species.append(spec_dict)
        # save products
        for i in range(sbml_reaction.getNumProducts()):
            spec = sbml_reaction.getProduct(i)
            spec_dict = {
                SBML_DFS.RSC_ID: spec.getId(),
                SBML_DFS.SC_ID: spec.getSpecies(),
                SBML_DFS.STOICHIOMETRY: spec.getStoichiometry(),
                SBML_DFS.SBO_TERM: spec.getSBOTermID(),
            }
            reaction_species.append(spec_dict)

        self.species = pd.DataFrame(reaction_species).set_index(SBML_DFS.RSC_ID)


def sbml_dfs_from_sbml(self, sbml_model: SBML, compartment_aliases: dict | None = None):
    """Parses an SBML model into a set of standardized DataFrames.

    This function serves as the main entry point for converting an SBML model
    into the internal DataFrame-based representation used by napistu. It
    orchestrates the processing of compartments, species, and reactions.

    Parameters
    ----------
    self : object
        The instance of the calling class, expected to have a `schema` attribute.
    sbml_model : SBML
        The SBML model to be parsed.
    compartment_aliases : dict, optional
        A dictionary to map custom compartment names to the napistu controlled
        vocabulary. If None, the default mapping (COMPARTMENT_ALIASES) is used.
        Defaults to None.

    Returns
    -------
    object
        The calling class instance, now populated with DataFrames for
        compartments, species, compartmentalized_species, reactions, and reaction_species
    """
    # 1. Process compartments from the SBML model
    self.compartments = sbml_model._define_compartments(compartment_aliases)

    # 2. Process species and compartmentalized species
    self.species, self.compartmentalized_species = sbml_model._define_species()

    # Refine compartments to only those actually used by species
    self.compartments = _refine_compartments(
        self.compartments, self.compartmentalized_species
    )

    # 3. Process reactions and their participating species
    self.reactions, self.reaction_species = sbml_model._define_reactions()

    return self


def _validate_species_consistency(species_df, compartmentalized_species_df):
    """Validate consistency between species and compartmentalized species tables.

    Parameters
    ----------
    species_df : pd.DataFrame
        DataFrame of species with s_id as index
    compartmentalized_species_df : pd.DataFrame
        DataFrame of compartmentalized species with s_id column

    Raises
    ------
    ValueError
        If there are inconsistencies between the two tables
    """
    species_ids_in_species = set(species_df.index)
    species_ids_in_cspecies = set(compartmentalized_species_df[SBML_DFS.S_ID])

    missing_in_species = species_ids_in_cspecies - species_ids_in_species
    missing_in_cspecies = species_ids_in_species - species_ids_in_cspecies

    if missing_in_species:
        raise ValueError(
            f"Species extraction validation error: {len(missing_in_species)} species IDs "
            f"found in compartmentalized_species but missing from species table. "
            f"Missing species: {sorted(list(missing_in_species))[:10]}"
        )

    if missing_in_cspecies:
        raise ValueError(
            f"Species extraction validation error: {len(missing_in_cspecies)} species IDs "
            f"found in species table but missing from compartmentalized_species table. "
            f"Missing from compartmentalized_species: {sorted(list(missing_in_cspecies))[:10]}"
        )


def _refine_compartments(compartments_df, compartmentalized_species_df):
    """Refine compartments to only those actually used by compartmentalized species.

    This function filters the compartments DataFrame to include only compartments
    that are referenced by compartmentalized species, and validates that all
    required compartments exist.

    Parameters
    ----------
    compartments_df : pd.DataFrame
        DataFrame of all extracted compartments with c_id as index
    compartmentalized_species_df : pd.DataFrame
        DataFrame of compartmentalized species with c_id column

    Returns
    -------
    pd.DataFrame
        Filtered compartments DataFrame containing only used compartments

    Raises
    ------
    ValueError
        If compartmentalized species reference compartments that don't exist
    """
    compartment_ids_in_compartments = set(compartments_df.index)
    compartment_ids_in_cspecies = set(compartmentalized_species_df[SBML_DFS.C_ID])
    # ignore None in compartment_ids_in_cspecies - this can likely be resolved downstream
    compartment_ids_in_cspecies = {
        c for c in compartment_ids_in_cspecies if c is not None
    }

    # Check that all required compartments exist
    missing_compartments = compartment_ids_in_cspecies - compartment_ids_in_compartments
    if missing_compartments:
        raise ValueError(
            f"Compartment refinement error: {len(missing_compartments)} compartment IDs "
            f"found in compartmentalized_species but missing from compartments table. "
            f"Missing compartments: {sorted(list(missing_compartments))[:10]}"
        )

    # Filter to only compartments that are actually used
    unused_compartments = compartment_ids_in_compartments - compartment_ids_in_cspecies
    if unused_compartments:
        logger.debug(
            f"Filtering out {len(unused_compartments)} unused compartments from compartments table. "
            f"Unused compartments: {sorted(list(unused_compartments))[:10]}"
        )

    # Return only the compartments that are referenced by species
    used_compartments = compartments_df.loc[
        compartments_df.index.isin(compartment_ids_in_cspecies)
    ]

    logger.debug(
        f"Refined compartments: {len(used_compartments)} of {len(compartments_df)} compartments retained"
    )

    return used_compartments


def _define_compartments_missing_cvterms(
    comp: libsbml.Compartment, aliases: dict
) -> dict[str, Any]:

    comp_name = comp.getName()
    mapped_compartment_key = [
        compkey for compkey, mappednames in aliases.items() if comp_name in mappednames
    ]

    if len(mapped_compartment_key) == 0:
        logger.warning(
            f"No GO compartment for {comp_name} is mapped, use the generic cellular_component's GO id"
        )

        compartment_entry = {
            SBML_DFS.C_ID: comp.getId(),
            SBML_DFS.C_NAME: comp.getName(),
            SBML_DFS.C_IDENTIFIERS: identifiers.Identifiers(
                [
                    identifiers.format_uri(
                        uri=identifiers.create_uri_url(
                            ontology=ONTOLOGIES.GO,
                            identifier=COMPARTMENTS_GO_TERMS[GENERIC_COMPARTMENT],
                        ),
                        bqb=BQB.BQB_IS,
                    )
                ]
            ),
            SBML_DFS.C_SOURCE: source.Source.empty(),
        }

    if len(mapped_compartment_key) > 0:
        if len(mapped_compartment_key) > 1:
            logger.warning(
                f"More than one GO compartments for {comp_name} are mapped, using the first one"
            )

        compartment_entry = {
            SBML_DFS.C_ID: comp.getId(),
            SBML_DFS.C_NAME: comp.getName(),
            SBML_DFS.C_IDENTIFIERS: identifiers.Identifiers(
                [
                    identifiers.format_uri(
                        uri=identifiers.create_uri_url(
                            ontology=ONTOLOGIES.GO,
                            identifier=COMPARTMENTS_GO_TERMS[mapped_compartment_key[0]],
                        ),
                        bqb=BQB.IS,
                    )
                ]
            ),
            SBML_DFS.C_SOURCE: source.Source.empty(),
        }

    return compartment_entry


def _get_gene_product_dict(gp):
    """Extracts attributes of a gene product from an SBML reaction object.

    Parameters
    ----------
    gp : libsbml.GeneProduct
        A libsbml GeneProduct object.

    Returns
    -------
    dict
        A dictionary containing the gene product's ID, name, and identifiers.
    """
    return {
        SBML_DFS.RSC_ID: gp.getId(),
        SBML_DFS.SC_ID: gp.getGeneProduct(),
        SBML_DFS.STOICHIOMETRY: 0,
        SBML_DFS.SBO_TERM: gp.getSBOTermID(),
    }


def _extract_gene_products(association: libsbml.Association) -> list[dict]:
    """Recursively extracts gene products from an association tree."""
    gene_products = []

    def _recursive_helper(assoc: libsbml.Association):
        if hasattr(assoc, SBML_DEFS.REACTION_ATTR_GET_GENE_PRODUCT):
            gene_products.append(_get_gene_product_dict(assoc))
        elif hasattr(assoc, "getNumAssociations"):
            for i in range(assoc.getNumAssociations()):
                _recursive_helper(assoc.getAssociation(i))

    _recursive_helper(association)
    return gene_products



from __future__ import annotations

import datetime
import logging
import os
from typing import Dict, Set, Tuple, Union

import igraph as ig
import numpy as np
import pandas as pd
import requests

from napistu import sbml_dfs_utils, utils
from napistu.constants import (
    BQB,
    IDENTIFIERS,
    ONTOLOGIES,
    ONTOLOGIES_LIST,
    SBML_DFS,
    SBOTERM_NAMES,
)
from napistu.identifiers import Identifiers, parse_ensembl_id
from napistu.ingestion.constants import (
    DATA_SOURCE_DESCRIPTIONS,
    DATA_SOURCES,
    DEFAULT_INTACT_RELATIVE_WEIGHTS,
    INTACT_EXPERIMENTAL_ROLES,
    INTACT_ONTOLOGY_ALIASES,
    INTACT_PUBLICATION_SCORE_THRESHOLD,
    INTACT_SCORES,
    INTACT_TERM_SCORES,
    INTERACTION_EDGELIST_DEFS,
    PSI_MI_DEFS,
    PSI_MI_INTACT_FTP_URL,
    PSI_MI_INTACT_SPECIES_TO_BASENAME,
    PSI_MI_MISSING_VALUE_STR,
    PSI_MI_ONTOLOGY_URL,
    PSI_MI_SCORED_TERMS,
    PSI_MI_STUDY_TABLES,
    PSI_MI_STUDY_TABLES_LIST,
    VALID_INTACT_EXPERIMENTAL_ROLES,  # noqa: F401
    VALID_INTACT_SECONDARY_ONTOLOGIES,
)
from napistu.ingestion.organismal_species import OrganismalSpeciesValidator
from napistu.ontologies import renaming
from napistu.sbml_dfs_core import SBML_dfs
from napistu.source import Source

logger = logging.getLogger(__name__)


def download_intact_xmls(
    output_dir_path: str,
    organismal_species: Union[str, OrganismalSpeciesValidator],
    overwrite: bool = False,
) -> None:
    """
    Download IntAct Species

    Download the PSM-30 XML files from IntAct for a species of interest.

    Parameters
    ----------
    output_dir_path : str
        Local directory to create an unzip files into
    latin_species : str
        The species name (e.g., "Homo sapiens") to work with
    overwrite : bool, optional
        Overwrite an existing output directory, by default False

    Returns
    -------
    None
        Files are downloaded and extracted to the specified directory
    """

    organismal_species = OrganismalSpeciesValidator.ensure(organismal_species)
    intact_species_basename = _get_intact_species_basename(
        organismal_species.latin_name
    )

    intact_species_url = os.path.join(
        PSI_MI_INTACT_FTP_URL, f"{intact_species_basename}.zip"
    )

    logger.info(f"Downloading and unzipping {intact_species_url}")

    utils.download_and_extract(
        intact_species_url,
        output_dir_path=output_dir_path,
        download_method="ftp",
        overwrite=overwrite,
    )


def intact_to_sbml_dfs(
    intact_summaries: dict[str, pd.DataFrame],
    organismal_species: Union[str, OrganismalSpeciesValidator],
) -> SBML_dfs:
    """
    Convert IntAct summaries to SBML_dfs

    Parameters
    ----------
    intact_summaries : dict[str, pd.DataFrame]
        A dictionary of IntAct summaries.
    organismal_species : str | OrganismalSpeciesValidator
        The organismal species pertaining to the IntAct interactions

    Returns
    -------
    sbml_dfs : SBML_dfs
        SBML_dfs object containing the converted IntAct data

    Raises
    ------
    ValueError
        If intact_summaries does not contain the required tables
    ValueError
        If the provided species is not supported by IntAct
    ValueError
        If ontologies listed as valid secondary references are not in the Napistu controlled vocabulary
    """

    organismal_species = OrganismalSpeciesValidator.ensure(organismal_species)

    if set(intact_summaries.keys()) != set(PSI_MI_STUDY_TABLES_LIST):
        raise ValueError(
            f"IntAct summaries must contain the following tables: {PSI_MI_STUDY_TABLES_LIST}"
        )

    # format model-level metadata
    model_source = Source.single_entry(
        model=DATA_SOURCES.INTACT,
        pathway_id=DATA_SOURCES.INTACT,
        data_source=DATA_SOURCES.INTACT,
        organismal_species=organismal_species.latin_name,
        name=DATA_SOURCE_DESCRIPTIONS[DATA_SOURCES.INTACT],
        date=datetime.date.today().strftime("%Y%m%d"),
    )

    aliases = renaming.OntologySet(ontologies=INTACT_ONTOLOGY_ALIASES).ontologies
    alias_mapping = renaming._create_alias_mapping(aliases)

    valid_intact_xrefs = _filter_intact_xrefs(
        intact_summaries,
        alias_mapping,
        organismal_species,
        VALID_INTACT_SECONDARY_ONTOLOGIES,
    )

    # filter to entries with valid xrefs
    valid_interactors = intact_summaries[PSI_MI_STUDY_TABLES.SPECIES].merge(
        valid_intact_xrefs[
            [PSI_MI_DEFS.STUDY_ID, PSI_MI_DEFS.INTERACTOR_ID]
        ].drop_duplicates(),
        how="inner",
    )

    lookup_table, species_df = _create_species_df(
        valid_interactors, valid_intact_xrefs, organismal_species.latin_name
    )

    # turn reaction_species into a bait <-> prey edgelist
    basic_edgelist = _create_basic_edgelist(intact_summaries, lookup_table)

    # drop species that are not in the edgelist
    defined_interactors = set(
        basic_edgelist[INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM]
    ) | set(basic_edgelist[INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM])
    used_species_mask = species_df[SBML_DFS.S_NAME].isin(defined_interactors)

    if len(species_df[~used_species_mask]) > 0:
        logger.warning(
            f"Dropping {sum(~used_species_mask)} species that are not in the edgelist"
        )
        species_df = species_df[used_species_mask]

    # add metadata
    edgelist_w_study_metadata = basic_edgelist.merge(
        intact_summaries[PSI_MI_STUDY_TABLES.STUDY_LEVEL_DATA][
            [
                PSI_MI_DEFS.STUDY_ID,
                PSI_MI_DEFS.INTERACTION_METHOD,
                IDENTIFIERS.ONTOLOGY,
                IDENTIFIERS.IDENTIFIER,
            ]
        ]
    )

    # drop entries no study id
    invalid_study_id_mask = edgelist_w_study_metadata[IDENTIFIERS.ONTOLOGY] == ""
    if sum(invalid_study_id_mask) > 0:
        logger.warning(
            f"Dropping {sum(invalid_study_id_mask)} interactions with no study id"
        )
        edgelist_w_study_metadata = edgelist_w_study_metadata[~invalid_study_id_mask]

    # convert from specified attributes to standardized attributes with associated scores
    standardized_interaction_attrs = _standardize_interaction_attrs(
        edgelist_w_study_metadata
    )
    # deduplicate and count the number of studies with scored attributes
    scored_attribute_counts = _count_studies_with_scored_attributes(
        standardized_interaction_attrs
    )

    # create r_Identifiers objects and count citations and attributes
    interaction_edgelist_df_ids_and_counts = _define_edgelist_df_ids_and_counts(
        edgelist_w_study_metadata, alias_mapping, scored_attribute_counts
    )

    # calculate publication, method, and type scores and weight them for the composite score
    interaction_scores = _calculate_all_scores_vectorized(
        scored_attribute_counts,
        interaction_edgelist_df_ids_and_counts[INTACT_SCORES.N_PUBLICATIONS],
    )

    interactions_edgelist = (
        edgelist_w_study_metadata[
            [
                INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
                INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
                PSI_MI_DEFS.INTERACTION_NAME,
            ]
        ]
        .rename(columns={PSI_MI_DEFS.INTERACTION_NAME: SBML_DFS.R_NAME})
        .groupby(
            [
                INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
                INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
            ]
        )
        .first()
        .reset_index()
        .merge(
            interaction_scores,
            on=[
                INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
                INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
            ],
            how="left",
        )
        .merge(
            interaction_edgelist_df_ids_and_counts,
            on=[
                INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
                INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
            ],
            how="left",
        )
    )

    sbml_dfs = SBML_dfs.from_edgelist(
        interactions_edgelist,
        species_df,
        compartments_df=sbml_dfs_utils.stub_compartments(),
        model_source=model_source,
        interaction_edgelist_defaults={
            INTERACTION_EDGELIST_DEFS.SBO_TERM_NAME_UPSTREAM: SBOTERM_NAMES.INTERACTOR,
            INTERACTION_EDGELIST_DEFS.SBO_TERM_NAME_DOWNSTREAM: SBOTERM_NAMES.INTERACTOR,
            SBML_DFS.R_ISREVERSIBLE: True,
        },
        keep_reactions_data=DATA_SOURCES.INTACT,
        force_edgelist_consistency=True,
    )

    return sbml_dfs


# =============================================================================
# UTILITY FUNCTIONS (alphabetically ordered)
# =============================================================================


def _build_psi_mi_ontology_graph(ontology_url: str = PSI_MI_ONTOLOGY_URL) -> ig.Graph:
    """Parse MI ontology JSON from URL and build igraph directed graph."""
    # Fetch JSON from URL
    response = requests.get(ontology_url)
    response.raise_for_status()
    ontology_json = response.json()

    # Collect all nodes and edges
    nodes = []
    edges = []

    def parse_node(node, parent_name=None):
        name = node["name"]
        nodes.append(name)

        # Add edge from parent to child (directed)
        if parent_name:
            edges.append((parent_name, name))

        # Recursively parse children
        for child in node.get("children", []):
            parse_node(child, parent_name=name)

    # Parse the root node
    parse_node(ontology_json)

    # Create igraph
    g = ig.Graph.TupleList(edges, directed=True)
    # Add any isolated nodes
    for node in set(nodes) - set(g.vs["name"]):
        g.add_vertex(node)

    return g


def _calculate_all_scores_vectorized(
    counts_df: pd.DataFrame,
    n_publications_series: pd.Series,
    max_pubs: int = INTACT_PUBLICATION_SCORE_THRESHOLD,
    weights: Dict[str, float] = DEFAULT_INTACT_RELATIVE_WEIGHTS,
) -> pd.DataFrame:
    """
    Calculate all MIscore components using vectorized operations.

    Note: This implementation follows the MIscore mathematical formulas from
    Villaveces et al. (2015), but has not been validated against published
    IntAct scores due to lack of detailed worked examples showing:
    - Specific interaction evidence (X studies of type Y using method Z)
    - The resulting component scores
    - The final MIscore

    Parameters
    ----------
    counts_df : pd.DataFrame
        DataFrame containing interaction counts and scores
    n_publications_series : pd.Series
        Series containing publication counts for each interaction
    max_pubs : int, optional
        Maximum publication threshold for scoring, by default INTACT_PUBLICATION_SCORE_THRESHOLD
    weights : dict[str, float], optional
        Dictionary of weights for different score components, by default DEFAULT_INTACT_RELATIVE_WEIGHTS

    Returns
    -------
    pd.DataFrame
        DataFrame containing all calculated scores for each interaction

    Raises
    ------
    ValueError
        If the weights dictionary does not contain the expected keys
    """

    expected_weighted_keys = DEFAULT_INTACT_RELATIVE_WEIGHTS.keys()
    if set(weights.keys()) != set(expected_weighted_keys):
        raise ValueError(
            f"The weights dictionary must contain the following keys: {expected_weighted_keys}"
        )

    # Get all unique interactions
    interactions = counts_df[
        [
            INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
            INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
        ]
    ].drop_duplicates()

    # Calculate publication scores from the Series
    pub_scores = n_publications_series.to_frame(
        INTACT_SCORES.N_PUBLICATIONS
    ).reset_index()
    pub_scores[INTACT_SCORES.PUBLICATION_SCORE] = np.where(
        pub_scores[INTACT_SCORES.N_PUBLICATIONS] > max_pubs,
        1.0,
        np.log(pub_scores[INTACT_SCORES.N_PUBLICATIONS] + 1) / np.log(max_pubs + 1),
    )
    pub_scores = pub_scores.drop(columns=[INTACT_SCORES.N_PUBLICATIONS])

    # Calculate method scores
    method_scores = _calculate_category_scores_vectorized(
        counts_df, PSI_MI_DEFS.INTERACTION_METHOD
    )

    # Calculate type scores
    type_scores = _calculate_category_scores_vectorized(
        counts_df, PSI_MI_DEFS.INTERACTION_TYPE
    )

    # Merge all scores together
    result = (
        interactions.merge(
            pub_scores,
            on=[
                INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
                INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
            ],
            how="left",
        )
        .merge(
            method_scores,
            on=[
                INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
                INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
            ],
            how="left",
        )
        .merge(
            type_scores,
            on=[
                INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
                INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
            ],
            how="left",
        )
        # Fill missing scores with 0
        .fillna(0.0)
    )

    # Calculate final MIscore
    total_weight = sum(weights.values())
    result[INTACT_SCORES.MI_SCORE] = (
        weights[INTACT_SCORES.PUBLICATION_SCORE]
        * result[INTACT_SCORES.PUBLICATION_SCORE]
        + weights[INTACT_SCORES.INTERACTION_METHOD_SCORE]
        * result[INTACT_SCORES.INTERACTION_METHOD_SCORE]
        + weights[INTACT_SCORES.INTERACTION_TYPE_SCORE]
        * result[INTACT_SCORES.INTERACTION_TYPE_SCORE]
    ) / total_weight

    return result


def _calculate_category_scores_vectorized(
    counts_df: pd.DataFrame, attribute_type: str
) -> pd.DataFrame:
    """
    Calculate method or type scores for all interactions using vectorized operations.

    Parameters
    ----------
    counts_df : pd.DataFrame
        DataFrame containing interaction data
    attribute_type : str
        Type of attribute to calculate scores for (e.g., 'interaction_method', 'interaction_type')

    Returns
    -------
    pd.DataFrame
        DataFrame containing category scores for each interaction
    """
    # Filter to specific attribute type
    filtered_df = counts_df[
        counts_df[INTACT_SCORES.ATTRIBUTE_TYPE] == attribute_type
    ].copy()

    if len(filtered_df) == 0:
        # Return empty dataframe with expected structure
        return pd.DataFrame(
            columns=[
                INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
                INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
                f"{attribute_type}_score",
            ]
        )

    # Calculate a = sum(score * count) for each interaction
    filtered_df["score_x_count"] = (
        filtered_df[INTACT_SCORES.RAW_SCORE] * filtered_df["count"]
    )
    a_values = (
        filtered_df.groupby(
            [
                INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
                INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
            ]
        )["score_x_count"]
        .sum()
        .reset_index()
        .rename(columns={"score_x_count": "a"})
    )

    # Calculate max count by score group for each interaction
    max_by_score_group = (
        filtered_df.groupby(
            [
                INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
                INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
                INTACT_SCORES.RAW_SCORE,
            ]
        )["count"]
        .max()
        .reset_index()
        .groupby(
            [
                INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
                INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
            ]
        )["count"]
        .sum()
        .reset_index()
        .rename(columns={"count": "max_sum"})
    )

    # Merge a and max_sum
    merged = a_values.merge(
        max_by_score_group,
        on=[
            INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
            INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
        ],
    )
    merged["b"] = merged["a"] + merged["max_sum"]

    # Calculate category score: log(a+1) / log(b+1)
    merged[f"{attribute_type}_score"] = np.where(
        merged["a"] == 0, 0.0, np.log(merged["a"] + 1) / np.log(merged["b"] + 1)
    )

    return merged[
        [
            INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
            INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
            f"{attribute_type}_score",
        ]
    ]


def _count_studies_with_scored_attributes(
    standardized_interaction_attrs: pd.DataFrame,
) -> pd.DataFrame:
    """
    Count the number of studies which report an interaction based on scored attributes.

    Parameters
    ----------
    standardized_interaction_attrs : pd.DataFrame
        Long-form dataframe with columns: upstream_name, downstream_name,
        study_id, attribute_type, scored_term, score

    Returns
    -------
    scored_attribute_counts : pd.DataFrame
        The number of studies and score for each interaction-attribute_type-scored_term combination.
    """

    scored_attribute_counts = (
        standardized_interaction_attrs[
            [
                INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
                INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
                PSI_MI_DEFS.STUDY_ID,
                INTACT_SCORES.ATTRIBUTE_TYPE,
                INTACT_SCORES.SCORED_TERM,
                INTACT_SCORES.RAW_SCORE,
            ]
        ]
        .drop_duplicates()
        # Count studies for each interaction-attribute_type-scored_term combination
        .groupby(
            [
                INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
                INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
                INTACT_SCORES.ATTRIBUTE_TYPE,
                INTACT_SCORES.SCORED_TERM,
                INTACT_SCORES.RAW_SCORE,  # Include score in groupby to preserve it
            ]
        )
        .size()
        .reset_index(name="count")
    )

    return scored_attribute_counts


def _create_basic_edgelist(
    intact_summaries: Dict[str, pd.DataFrame], lookup_table: pd.Series
) -> pd.DataFrame:
    """
    Create a basic edgelist from the IntAct summaries and lookup table.

    The edgelist is created by merging the IntAct summaries and lookup table on the study id and interaction id.
    The edgelist is then filtered to only include interactions where the bait is present and the prey is present.
    The edgelist is then pivoted based on the hub and spoke model used by IntAct, where a single bait is connected
    to one or more prey.

    Parameters
    ----------
    intact_summaries : Dict[str, pd.DataFrame]
        A dictionary of IntAct summaries, keyed by study id.
    lookup_table : pd.Series
        A lookup table of interaction ids and their corresponding interaction names.

    Returns
    -------
    edgelist_df : pd.DataFrame
        A dataframe of the edgelist with the following columns:
        - upstream_name : the name of the upstream node
        - downstream_name : the name of the downstream node
        - interaction_name : the name of the interaction
        - study_id : the id of the study
        - interaction_type : the type of interaction

    Notes
    -----
    The convention of associating each bait to many prey follows conventions set by
    yeast 2 hybrid screens but it is applied across the board even for technologies
    when bait-prey relationships are not appropriate (e.g., purifying a whole complex).
    In these cases IntAct chooses a random component to serve as the prey. This will make
    it more closely related in a network sense than its interactors than they would be
    to one another. This could be addressed by expanding interactions but this would  be
    quite tricky because some interactions have 100s of prey and (N choose 2) would be
    cumbersome. This could be done for just certain types of annotations but it seems
    like a big headache for very little practical gain.
    """

    interactions = (
        intact_summaries[PSI_MI_STUDY_TABLES.REACTION_SPECIES]
        .merge(
            lookup_table,
            on=[PSI_MI_DEFS.STUDY_ID, PSI_MI_DEFS.INTERACTOR_ID],
            how="inner",
        )
        .query("experimental_role in @VALID_INTACT_EXPERIMENTAL_ROLES")
    )

    # the hub and spoke model connects a single bait to 1 or more prey
    # following the conventions yeast 2 hybrid screens
    valid_interactions_structure = (
        interactions.value_counts(
            [
                PSI_MI_DEFS.STUDY_ID,
                PSI_MI_DEFS.INTERACTION_NAME,
                PSI_MI_DEFS.EXPERIMENTAL_ROLE,
            ]
        )
        .to_frame()
        .pivot_table(
            index=[PSI_MI_DEFS.STUDY_ID, PSI_MI_DEFS.INTERACTION_NAME],
            columns=PSI_MI_DEFS.EXPERIMENTAL_ROLE,
            values="count",  # or whatever your count column is called
            fill_value=0,  # fill missing values with 0
        )
        .query(
            f"{INTACT_EXPERIMENTAL_ROLES.PREY} >= 1 & {INTACT_EXPERIMENTAL_ROLES.BAIT} == 1"
        )
    )

    valid_reaction_species = interactions.merge(
        valid_interactions_structure,
        left_on=[PSI_MI_DEFS.STUDY_ID, PSI_MI_DEFS.INTERACTION_NAME],
        right_index=True,
    ).drop(columns=[INTACT_EXPERIMENTAL_ROLES.BAIT, INTACT_EXPERIMENTAL_ROLES.PREY])

    all_prey = (
        valid_reaction_species.query(
            "experimental_role == @INTACT_EXPERIMENTAL_ROLES.PREY"
        ).rename(columns={SBML_DFS.S_NAME: INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM})
    )[
        [
            PSI_MI_DEFS.STUDY_ID,
            PSI_MI_DEFS.INTERACTION_NAME,
            INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
        ]
    ]

    all_bait = (
        valid_reaction_species.query(
            "experimental_role == @INTACT_EXPERIMENTAL_ROLES.BAIT"
        ).rename(columns={SBML_DFS.S_NAME: INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM})
    )[
        [
            PSI_MI_DEFS.STUDY_ID,
            PSI_MI_DEFS.INTERACTION_NAME,
            PSI_MI_DEFS.INTERACTION_TYPE,
            INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
        ]
    ]

    edgelist_df = all_prey.merge(
        all_bait, on=[PSI_MI_DEFS.STUDY_ID, PSI_MI_DEFS.INTERACTION_NAME], how="left"
    )
    edgelist_df = edgelist_df[
        [
            INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
            INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
            PSI_MI_DEFS.INTERACTION_NAME,
            PSI_MI_DEFS.STUDY_ID,
            PSI_MI_DEFS.INTERACTION_TYPE,
        ]
    ]

    # flip upstream and downstream names if needed so
    # upstream name alphanumerically comes before downstream name

    from_orig = edgelist_df[INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM].copy()
    to_orig = edgelist_df[INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM].copy()

    # Create mask and assign
    needs_flip = from_orig > to_orig
    edgelist_df[INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM] = np.where(
        needs_flip, to_orig, from_orig
    )
    edgelist_df[INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM] = np.where(
        needs_flip, from_orig, to_orig
    )

    return edgelist_df


def _create_r_identifiers(group_data: pd.DataFrame) -> Identifiers:
    """
    Create a list of identifiers for an experiment's interactions.

    Parameters
    ----------
    group_data : pd.DataFrame
        A dataframe with the study metadata.

    Returns
    -------
    Identifiers
        An Identifiers object containing the interaction identifiers
    """

    interaction_dict = [
        {
            IDENTIFIERS.ONTOLOGY: x[IDENTIFIERS.ONTOLOGY],
            IDENTIFIERS.IDENTIFIER: x[IDENTIFIERS.IDENTIFIER],
            IDENTIFIERS.BQB: BQB.IS_DESCRIBED_BY,
        }
        for x in group_data.to_dict("records")
    ]

    return Identifiers(interaction_dict)


def _create_species_df(
    raw_species_df: pd.DataFrame,
    raw_species_identifiers_df: pd.DataFrame,
    organismal_species: Union[str, OrganismalSpeciesValidator],
) -> Tuple[pd.Series, pd.DataFrame]:
    """
    Create a species dataframe from the raw species dataframe and the raw species identifiers dataframe.

    Parameters
    ----------
    raw_species_df : pd.DataFrame
        The raw species dataframe.
    raw_species_identifiers_df : pd.DataFrame
        The raw species identifiers dataframe.
    organismal_species : str | OrganismalSpeciesValidator
        The organismal species pertaining to the IntAct interactions

    Returns
    -------
    lookup_table : pd.Series
        A lookup table mapping study_id and interactor_id to the molecular species name.
    species_df : pd.DataFrame
        The molecular species dataframe.

    Raises
    ------
    ValueError
        If the provided species is not supported by IntAct
    """

    organismal_species = OrganismalSpeciesValidator.ensure(organismal_species)
    intact_species_basename = _get_intact_species_basename(
        organismal_species.latin_name
    )

    raw_species_df = raw_species_df.copy()
    raw_species_df["species_match"] = raw_species_df[
        PSI_MI_DEFS.INTERACTOR_LABEL
    ].str.endswith(intact_species_basename)

    all_species_annotations = raw_species_identifiers_df.merge(
        raw_species_df[
            [
                PSI_MI_DEFS.INTERACTOR_ID,
                PSI_MI_DEFS.STUDY_ID,
                PSI_MI_DEFS.INTERACTOR_LABEL,
                "species_match",
            ]
        ],
        on=[PSI_MI_DEFS.STUDY_ID, PSI_MI_DEFS.INTERACTOR_ID],
        how="inner",
    )[
        [
            PSI_MI_DEFS.INTERACTOR_LABEL,
            IDENTIFIERS.ONTOLOGY,
            IDENTIFIERS.IDENTIFIER,
            IDENTIFIERS.BQB,
            "species_match",
            "white_listed",
        ]
    ].drop_duplicates()

    valid_species_annotations = (
        all_species_annotations
        # remove names which are not associated with at least 1 species match by name or a white-listed
        # annotation like a ChEBI id or a species-matched ensembl id
        .groupby("interactor_label")
        .filter(lambda group: (group["species_match"] | group["white_listed"]).any())
        .drop(columns=["species_match", "white_listed"])
    )

    # aggregate annotations so there is just 1 row per interactor label
    species_df = (
        valid_species_annotations
        # group by interactor label and apply the Identifiers constructor
        .groupby([PSI_MI_DEFS.INTERACTOR_LABEL])
        .apply(lambda x: Identifiers(x.to_dict(orient="records")), include_groups=False)
        .rename(SBML_DFS.S_IDENTIFIERS)
        .rename_axis(SBML_DFS.S_NAME)
        .reset_index()
    )

    lookup_table = (
        raw_species_df.loc[
            raw_species_df[PSI_MI_DEFS.INTERACTOR_LABEL].isin(
                species_df[SBML_DFS.S_NAME]
            )
        ]
        .rename(columns={PSI_MI_DEFS.INTERACTOR_LABEL: SBML_DFS.S_NAME})
        .set_index([PSI_MI_DEFS.STUDY_ID, PSI_MI_DEFS.INTERACTOR_ID])[SBML_DFS.S_NAME]
    )

    return lookup_table, species_df


def _define_edgelist_df_ids_and_counts(
    edgelist_w_study_metadata: pd.DataFrame,
    alias_mapping: Dict[str, str],
    scored_attribute_counts: pd.DataFrame,
) -> pd.DataFrame:
    """
    Add attributes to the edgelist.

    Parameters
    ----------
    edgelist_w_study_metadata : pd.DataFrame
        The edgelist with study metadata.
    alias_mapping : dict[str, str]
        A dictionary mapping from ontology aliases to the Napistu controlled vocabulary.
    scored_attribute_counts : pd.DataFrame
        A dataframe with the number of studies with each attribute.

    Returns
    -------
    pd.DataFrame
        A dataframe with the edgelist, citation counts, and identifiers.
    """

    # rename ontologies
    tmp_edgelist_w_study_metadata = edgelist_w_study_metadata.copy()
    tmp_edgelist_w_study_metadata[IDENTIFIERS.ONTOLOGY] = tmp_edgelist_w_study_metadata[
        IDENTIFIERS.ONTOLOGY
    ].replace(alias_mapping)

    unrecognized_ontologies = set(
        tmp_edgelist_w_study_metadata[IDENTIFIERS.ONTOLOGY]
    ) - set(ONTOLOGIES_LIST)

    if len(unrecognized_ontologies) > 0:
        logger.warning(
            f"The following ontologies were primary references for interactions but they are not part of the  Napistu controlled vocabulary (ONTOLOGIES): {unrecognized_ontologies}"
        )

    track_citations = edgelist_w_study_metadata[
        [
            INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
            INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
            PSI_MI_DEFS.STUDY_ID,
            IDENTIFIERS.ONTOLOGY,
            IDENTIFIERS.IDENTIFIER,
        ]
    ].drop_duplicates()

    edgelist_groups = track_citations.groupby(
        [
            INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
            INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
        ]
    )

    # count the # of citations
    edgelist_citation_counts = edgelist_groups.size().rename(
        INTACT_SCORES.N_PUBLICATIONS
    )
    edgelist_identifiers = edgelist_groups.apply(_create_r_identifiers).rename(
        SBML_DFS.R_IDENTIFIERS
    )

    # create a wide output with one column per interaction attribute
    tmp_scored_attribute_counts = scored_attribute_counts.copy()
    tmp_scored_attribute_counts["composite_attribute"] = (
        tmp_scored_attribute_counts["attribute_type"]
        + "_"
        + tmp_scored_attribute_counts["scored_term"]
    )
    wide_attribute_counts = tmp_scored_attribute_counts.pivot_table(
        index=[
            INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
            INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
        ],
        columns="composite_attribute",
        values="count",
        aggfunc="sum",
        fill_value=0,
    )

    return pd.concat(
        [edgelist_citation_counts, edgelist_identifiers, wide_attribute_counts], axis=1
    )


def _filter_intact_xrefs(
    intact_summaries: Dict[str, pd.DataFrame],
    alias_mapping: Dict[str, str],
    organismal_species: Union[str, OrganismalSpeciesValidator],
    valid_secondary_ontologies: Set[str] = VALID_INTACT_SECONDARY_ONTOLOGIES,
) -> pd.DataFrame:
    """
    Filter IntAct species identifiers to only those which should be added as s_Identifiers.

    Parameters
    ----------
    intact_summaries : Dict[str, pd.DataFrame]
        The IntAct summaries table.
    alias_mapping : Dict[str, str]
        A dictionary mapping from ontology aliases to the Napistu controlled vocabulary.
    organismal_species : str | OrganismalSpeciesValidator
        The organismal species pertaining to the IntAct interactions
    valid_secondary_ontologies : Set[str], optional
        A set of ontologies which are valid secondary references, by default VALID_INTACT_SECONDARY_ONTOLOGIES

    Returns
    -------
    pd.DataFrame
        A DataFrame of IntAct species identifiers which should be added as s_Identifiers.

    Raises
    ------
    ValueError
        If ontologies listed as valid secondary references are not in the Napistu controlled vocabulary
    """

    organismal_species = OrganismalSpeciesValidator.ensure(organismal_species)

    invalid_valid_secondary_ontologies = set(valid_secondary_ontologies) - set(
        ONTOLOGIES_LIST
    )
    if len(invalid_valid_secondary_ontologies) > 0:
        raise ValueError(
            f"The following ontologies are listed as valid secondary references but they are not in the Napistu controlled vocabulary (ONTOLOGIES_LIST): {invalid_valid_secondary_ontologies}"
        )

    # pull out the raw species identifiers table
    standardized_intact_xrefs = intact_summaries[
        PSI_MI_STUDY_TABLES.SPECIES_IDENTIFIERS
    ]

    # drop entries which are missing either their ontology or identifiers
    valid_xref_mask = (
        standardized_intact_xrefs[IDENTIFIERS.ONTOLOGY] != PSI_MI_MISSING_VALUE_STR
    ) & (standardized_intact_xrefs[IDENTIFIERS.IDENTIFIER] != PSI_MI_MISSING_VALUE_STR)

    if sum(~valid_xref_mask) > 0:
        logger.warning(
            f"Dropping {sum(~valid_xref_mask)} species identifiers which are missing either their ontology or identifier."
        )

    standardized_intact_xrefs = standardized_intact_xrefs[valid_xref_mask]

    # convert from general "ensembl" ontology to ensembl gene, protein, etc. and remove version numbers
    sanitized_intact_xrefs = _sanitize_identifiers(
        standardized_intact_xrefs, organismal_species
    )

    # convert ontologies into the Napistu controlled vocabulary
    sanitized_intact_xrefs[IDENTIFIERS.ONTOLOGY] = sanitized_intact_xrefs[
        IDENTIFIERS.ONTOLOGY
    ].map(lambda x: alias_mapping.get(x, x))

    # format primary references
    primary_refs = sanitized_intact_xrefs[
        sanitized_intact_xrefs[PSI_MI_DEFS.REF_TYPE] == PSI_MI_DEFS.PRIMARY
    ]
    # all entries in primary refs should have a valid ontology
    invalid_ontologies = set(primary_refs[IDENTIFIERS.ONTOLOGY]) - set(ONTOLOGIES_LIST)
    if len(invalid_ontologies) > 0:
        _log_invalid_primary_refs(primary_refs, invalid_ontologies)
        primary_refs = primary_refs[
            ~primary_refs[IDENTIFIERS.ONTOLOGY].isin(invalid_ontologies)
        ]

    # only look at a subset of the secondary references
    # including additional ontologies to this set results in (rediculous) overmerging of species
    # when creating consensus models where species are defined by their s_Identifiers
    secondary_refs = sanitized_intact_xrefs[
        sanitized_intact_xrefs[PSI_MI_DEFS.REF_TYPE] == PSI_MI_DEFS.SECONDARY
    ].query("ontology in @VALID_INTACT_SECONDARY_ONTOLOGIES")
    # drop entries which are missing their primary ref
    secondary_refs = secondary_refs.merge(
        primary_refs[[PSI_MI_DEFS.STUDY_ID, PSI_MI_DEFS.INTERACTOR_ID]], how="inner"
    )

    valid_xrefs = pd.concat([primary_refs, secondary_refs]).assign(
        **{IDENTIFIERS.BQB: BQB.IS}
    )

    return valid_xrefs


def _get_intact_scored_term(term_name: str, term_lookup: Dict[str, str]) -> str:
    """Get the scored ancestor term name for a given term."""
    return term_lookup.get(term_name, PSI_MI_SCORED_TERMS.UNKNOWN)


def _get_intact_species_basename(latin_species: str) -> str:

    if latin_species not in PSI_MI_INTACT_SPECIES_TO_BASENAME.keys():
        raise ValueError(
            f"The provided species {latin_species} did not match any of the species in INTACT_SPECIES_TO_BASENAME: "
            f"{', '.join(PSI_MI_INTACT_SPECIES_TO_BASENAME.keys())}"
            "If this is a species supported by IntAct please add the species to the PSI_MI_INTACT_SPECIES_TO_BASENAME dictionary."
        )

    return PSI_MI_INTACT_SPECIES_TO_BASENAME[latin_species]


def _get_intact_term_with_score(
    ontology_graph: ig.Graph, scored_terms: Dict[str, float]
) -> Dict[str, str]:
    """Build lookup mapping all terms to their scored ancestor names."""
    term_lookup = {}

    for vertex in ontology_graph.vs:
        term_name = vertex["name"]

        # Check if term has explicit score
        if term_name in scored_terms:
            term_lookup[term_name] = term_name  # Maps to itself
        else:
            # Get all ancestors (nodes reachable via incoming edges)
            ancestors_idx = ontology_graph.subcomponent(vertex.index, mode="in")
            ancestors = [
                ontology_graph.vs[idx]["name"]
                for idx in ancestors_idx
                if idx != vertex.index
            ]

            # Find highest scoring ancestor
            scored_ancestors = [
                (ancestor, scored_terms[ancestor])
                for ancestor in ancestors
                if ancestor in scored_terms
            ]

            if scored_ancestors:
                # Map to the name of the highest scoring ancestor
                best_ancestor = max(scored_ancestors, key=lambda x: x[1])[0]
                term_lookup[term_name] = best_ancestor

    return term_lookup


def _log_invalid_primary_refs(
    primary_refs: pd.DataFrame, invalid_ontologies: Set[str]
) -> None:

    invalid_ontologies_counts = primary_refs[
        primary_refs[IDENTIFIERS.ONTOLOGY].isin(invalid_ontologies)
    ].value_counts(IDENTIFIERS.ONTOLOGY)

    invalid_ontologies_str = ", ".join(
        [
            f"{ontology} (N={count})"
            for ontology, count in invalid_ontologies_counts.items()
        ]
    )

    logger.warning(
        "The following ontologies are listed as primary species references but they are not\n"
        "in the Napistu controlled vocabulary (ONTOLOGIES_LIST). These species and their interactions will be ignored:\n"
        f"{invalid_ontologies_str}"
    )


def _process_ensembl_ids(ensembl_primary_xrefs, latin_species: str) -> pd.DataFrame:
    """
    Process ensembl IntAct references to convert them from the meta ensembl ontology to the Napistu controlled vocabulary.

    Parameters
    ----------
    ensembl_primary_xrefs : pd.DataFrame
        The standardized IntAct references filtered to the "ensembl" ontology.
    latin_species : str
        The latin species name to filter ensembl ids to.

    Returns
    -------
    pd.DataFrame
        The processed ensembl IntAct references.
    """

    def _safe_parse_ensembl_id(x):
        try:
            return parse_ensembl_id(x)
        except (TypeError, AttributeError):
            return (None, None, None)

    distinct_ensembl_ids = (
        ensembl_primary_xrefs[IDENTIFIERS.IDENTIFIER]
        .to_frame()
        .drop_duplicates()
        .rename(columns={IDENTIFIERS.IDENTIFIER: "original_identifier"})
    )

    distinct_ensembl_ids[
        [IDENTIFIERS.IDENTIFIER, IDENTIFIERS.ONTOLOGY, "latin_species"]
    ] = distinct_ensembl_ids["original_identifier"].apply(
        lambda x: pd.Series(_safe_parse_ensembl_id(x))
    )

    invalid_ensembl_ids = distinct_ensembl_ids[
        distinct_ensembl_ids[IDENTIFIERS.IDENTIFIER].isna()
    ]
    if len(invalid_ensembl_ids) > 0:
        logger.warning(
            f"Dropped {len(invalid_ensembl_ids)} ensembl ids which could not be formatted by identifiers.parse_ensembl_id: {invalid_ensembl_ids['original_identifier'].tolist()}"
        )

    distinct_ensembl_ids = distinct_ensembl_ids.dropna(subset=[IDENTIFIERS.IDENTIFIER])

    # filter by latin name
    correct_species_mask = distinct_ensembl_ids["latin_species"] == latin_species
    if sum(~correct_species_mask) > 0:
        logger.warning(
            f"Dropped {sum(~correct_species_mask)} ensembl ids which did not match the specified latin species name {latin_species} and will be ignored."
        )

    distinct_ensembl_ids = distinct_ensembl_ids[correct_species_mask]

    processed_ensembl_ids = (
        ensembl_primary_xrefs.rename(
            columns={IDENTIFIERS.IDENTIFIER: "original_identifier"}
        )
        .drop(columns=[IDENTIFIERS.ONTOLOGY])
        .merge(distinct_ensembl_ids, on="original_identifier", how="inner")
        .drop(columns=["original_identifier", "latin_species"])
    )

    return processed_ensembl_ids


def _sanitize_identifiers(
    standardized_intact_xrefs,
    organismal_species: str | OrganismalSpeciesValidator,
    ensembl_ontology_name: str = "ensembl",
    chebi_ontology_name: str = ONTOLOGIES.CHEBI,
    rna_central_ontology_name: str = ONTOLOGIES.RNACENTRAL,
) -> pd.DataFrame:
    """
    Sanitizes the identifiers in the standardized IntAct references.

    This functions applies ontology-specific manipulations and white lists non-genic molecular species like metabolites so they aren't filtered downstream.

    Parameters
    ----------
    standardized_intact_xrefs : pd.DataFrame
        The standardized IntAct references.
    organismal_species : str
        The organismal species to filter ensembl ids to.
    ensembl_ontology_name : str
        The name of the ontology to convert from.
    chebi_ontology_name : str
        The name of the ontology to convert from.
    rna_central_ontology_name : str
        The name of the ontology to convert from.

    Returns
    -------
    pd.DataFrame
        The sanitized IntAct references.
    """

    organismal_species = OrganismalSpeciesValidator.ensure(organismal_species)

    # split ontologies needing special treatmnet
    ensembl_primary_xrefs_mask = (
        standardized_intact_xrefs[PSI_MI_DEFS.REF_TYPE] == PSI_MI_DEFS.PRIMARY
    ) & (standardized_intact_xrefs[IDENTIFIERS.ONTOLOGY] == ensembl_ontology_name)
    if sum(ensembl_primary_xrefs_mask) == 0:
        logger.warning(
            "No ensembl primary references found in the standardized IntAct references. This may indicate a change in IntAct notation."
        )

    chebi_xrefs_mask = (
        standardized_intact_xrefs[IDENTIFIERS.ONTOLOGY] == chebi_ontology_name
    )
    if sum(chebi_xrefs_mask) == 0:
        logger.warning(
            "No CHEBI references found in the standardized IntAct references. This may indicate a change in IntAct notation."
        )

    rna_central_mask = (
        standardized_intact_xrefs[IDENTIFIERS.ONTOLOGY] == rna_central_ontology_name
    )
    if sum(rna_central_mask) == 0:
        logger.warning(
            "No RNA Central references found in the standardized IntAct references. This may indicate a change in IntAct notation."
        )

    other_ontologies_mask = ~(
        ensembl_primary_xrefs_mask | chebi_xrefs_mask | rna_central_mask
    )

    # format special ontologies
    ensembl_ids = _process_ensembl_ids(
        standardized_intact_xrefs.loc[ensembl_primary_xrefs_mask],
        organismal_species.latin_name,
    )
    chebi_ids = standardized_intact_xrefs.loc[chebi_xrefs_mask]
    # remove the "CHEBI:" prefix
    chebi_ids.loc[:, IDENTIFIERS.IDENTIFIER] = chebi_ids[
        IDENTIFIERS.IDENTIFIER
    ].str.replace("CHEBI:", "")

    white_listed_ids = pd.concat(
        [chebi_ids, ensembl_ids, standardized_intact_xrefs.loc[rna_central_mask]]
    ).assign(white_listed=True)

    return pd.concat(
        [
            white_listed_ids,
            standardized_intact_xrefs.loc[other_ontologies_mask].assign(
                white_listed=False
            ),
        ]
    )


def _standardize_interaction_attrs(
    edgelist_w_study_metadata: pd.DataFrame,
    ontology_url: str = PSI_MI_ONTOLOGY_URL,
) -> pd.DataFrame:

    # read ontology hierarchy from the IntAct github
    psi_mi_ontology_graph = _build_psi_mi_ontology_graph(ontology_url)
    # map from recognized terms in the ontology to terms with an explicit score
    # (other entries will be assigned unknown and receive its score)
    intact_term_lookup = _get_intact_term_with_score(
        psi_mi_ontology_graph, INTACT_TERM_SCORES
    )

    tall_interaction_attrs = pd.melt(
        edgelist_w_study_metadata,
        id_vars=[
            INTERACTION_EDGELIST_DEFS.NAME_UPSTREAM,
            INTERACTION_EDGELIST_DEFS.NAME_DOWNSTREAM,
            PSI_MI_DEFS.INTERACTION_NAME,
            PSI_MI_DEFS.STUDY_ID,
        ],
        value_vars=[PSI_MI_DEFS.INTERACTION_TYPE, PSI_MI_DEFS.INTERACTION_METHOD],
        var_name=INTACT_SCORES.ATTRIBUTE_TYPE,
        value_name=INTACT_SCORES.ATTRIBUTE_VALUE,
    )

    attribute_scores = (
        tall_interaction_attrs[INTACT_SCORES.ATTRIBUTE_VALUE]
        .drop_duplicates()
        .reset_index(drop=True)
        .to_frame()
    )

    # map to an appropraite term with a score
    attribute_scores[INTACT_SCORES.SCORED_TERM] = attribute_scores[
        INTACT_SCORES.ATTRIBUTE_VALUE
    ].apply(lambda x: _get_intact_scored_term(x, intact_term_lookup))
    attribute_scores[INTACT_SCORES.RAW_SCORE] = attribute_scores[
        INTACT_SCORES.SCORED_TERM
    ].apply(lambda x: INTACT_TERM_SCORES[x])

    tall_standardized_interaction_attrs = tall_interaction_attrs.merge(
        attribute_scores, on=INTACT_SCORES.ATTRIBUTE_VALUE, how="left"
    )

    return tall_standardized_interaction_attrs



"""
Utilities for loading and processing documentation.
"""

import os
from typing import Any, Dict

import httpx

from napistu.constants import PACKAGE_DEFS
from napistu.mcp import utils as mcp_utils
from napistu.mcp.constants import (
    DEFAULT_GITHUB_API,
    DOCUMENTATION,
    GITHUB_DEFS,
    GITHUB_ISSUES_INDEXED,
    GITHUB_PRS_INDEXED,
    SEARCH_RESULT_DEFS,
    SEARCH_TYPES,
)


async def fetch_wiki_page(
    page_name: str,
    repo: str = PACKAGE_DEFS.GITHUB_PROJECT_REPO,
    owner: str = PACKAGE_DEFS.GITHUB_OWNER,
) -> str:
    """
    Fetch wiki page content using raw GitHub URLs.

    Parameters
    ----------
    page_name : str
        The name of the page (without .md extension for wiki pages)
    repo : str
        The repository name
    owner : str
        The GitHub username or organization

    Returns
    -------
    str
        The raw Markdown content
    """
    # Use raw.githubusercontent.com for wiki pages
    url = f"https://raw.githubusercontent.com/wiki/{owner}/{repo}/{page_name}.md"

    async with httpx.AsyncClient() as client:
        resp = await client.get(url)
        resp.raise_for_status()
        return resp.text


async def get_issue(
    repo: str,
    number: int,
    owner: str = PACKAGE_DEFS.GITHUB_OWNER,
    github_api: str = DEFAULT_GITHUB_API,
) -> dict:
    """
    Get a single issue (or PR) by number from a GitHub repository.

    Parameters
    ----------
    repo : str
        The repository name.
    number : int
        The issue or PR number.
    owner : str, optional
        The GitHub username or organization (default is 'napistu').
    github_api : str, optional
        The GitHub API base URL (default is 'https://api.github.com').

    Returns
    -------
    dict
        The issue or PR details as a dictionary.
    """
    url = f"{github_api}/repos/{owner}/{repo}/issues/{number}"
    async with httpx.AsyncClient(headers=_get_github_headers()) as client:
        resp = await client.get(url)
        resp.raise_for_status()
        item = resp.json()
        return _format_github_issue(item)


async def list_issues(
    repo: str,
    owner: str = PACKAGE_DEFS.GITHUB_OWNER,
    github_api: str = DEFAULT_GITHUB_API,
    state: str = GITHUB_ISSUES_INDEXED,
    include_prs: bool = False,
) -> list:
    """
    List issues (and optionally PRs) for a given GitHub repository using the GitHub API.

    Parameters
    ----------
    repo : str, optional
        The repository name.
    owner : str, optional
        The GitHub username or organization (default is 'napistu').
    github_api : str, optional
        The GitHub API base URL (default is 'https://api.github.com').
    state : str, optional
        The state of the issues to return. Can be 'open', 'closed', or 'all'. Default is 'open'.
    include_prs : bool, optional
        If True, include pull requests in the results. Default is False.

    Returns
    -------
    list of dict
        Each dict contains: number, title, state, url, and a truncated body (max 500 chars).
    """
    url = f"{github_api}/repos/{owner}/{repo}/issues?state={state}"
    filter_func = (
        (lambda item: True)
        if include_prs
        else (lambda item: GITHUB_DEFS.PULL_REQUEST not in item)
    )
    return await _fetch_github_items(url, filter_func=filter_func)


async def list_pull_requests(
    repo: str,
    owner: str = PACKAGE_DEFS.GITHUB_OWNER,
    github_api: str = DEFAULT_GITHUB_API,
    state: str = GITHUB_PRS_INDEXED,
) -> list:
    """
    List pull requests for a given GitHub repository using the GitHub API.

    Parameters
    ----------
    repo : str, optional
        The repository name.
    owner : str, optional
        The GitHub username or organization (default is 'napistu').
    github_api : str, optional
        The GitHub API base URL (default is 'https://api.github.com').
    state : str, optional
        The state of the PRs to return. Can be 'open', 'closed', or 'all'. Default is 'open'.

    Returns
    -------
    list of dict
        Each dict contains: number, title, state, url, and a truncated body (max 500 chars).
    """
    url = f"{github_api}/repos/{owner}/{repo}/pulls?state={state}"
    return await _fetch_github_items(url)


async def load_readme_content(readme_url: str) -> str:
    if readme_url.startswith("http://") or readme_url.startswith("https://"):
        async with httpx.AsyncClient() as client:
            response = await client.get(readme_url)
            response.raise_for_status()
            return response.text
    else:
        raise ValueError(
            f"Only HTTP(S) URLs are supported for documentation paths: {readme_url}"
        )


# private utils


def _exact_search_documentation(
    query: str, docs_cache: Dict[str, Dict[str, Any]], max_exact_results: int = 20
) -> Dict[str, Any]:
    """
    Perform exact text search across documentation content.

    Searches for the query string in README files, wiki pages, GitHub issues,
    and pull requests. Returns results organized by type, or an error if
    too many matches are found.

    Parameters
    ----------
    query : str
        Search query string to match against documentation content
    docs_cache : Dict[str, Dict[str, Any]]
        Documentation cache dictionary with keys:
        - DOCUMENTATION.README: Dictionary of README content
        - DOCUMENTATION.WIKI: Dictionary of wiki page content
        - DOCUMENTATION.ISSUES: Dictionary mapping repos to lists of issues
        - DOCUMENTATION.PRS: Dictionary mapping repos to lists of PRs
    max_exact_results : int, optional
        Maximum number of total results allowed before returning an error.
        Default is 20.

    Returns
    -------
    Dict[str, Any]
        Search results dictionary containing:
        - query : str
            Original search query
        - search_type : str
            Always "exact"
        - results : Dict[str, List]
            Dictionary with keys README, WIKI, ISSUES, PRS, each containing
            a list of matching items with name, snippet, title, and url
        - tip : str
            Helpful guidance for improving search results
        OR (if too many results):
        - error : str
            Error message indicating too many results
        - suggestion : str
            Suggestions for refining the search
        - result_counts : Dict[str, int]
            Counts of matches by type
    """
    results = {
        DOCUMENTATION.README: [],
        DOCUMENTATION.WIKI: [],
        DOCUMENTATION.ISSUES: [],
        DOCUMENTATION.PRS: [],
    }

    # Search README files
    for readme_name, content in docs_cache[DOCUMENTATION.README].items():
        if query.lower() in content.lower():
            results[DOCUMENTATION.README].append(
                {
                    SEARCH_RESULT_DEFS.NAME: readme_name,
                    SEARCH_RESULT_DEFS.SNIPPET: mcp_utils.get_snippet(content, query),
                }
            )

    # Search wiki pages
    for page_name, content in docs_cache[DOCUMENTATION.WIKI].items():
        if query.lower() in content.lower():
            results[DOCUMENTATION.WIKI].append(
                {
                    SEARCH_RESULT_DEFS.NAME: page_name,
                    SEARCH_RESULT_DEFS.SNIPPET: mcp_utils.get_snippet(content, query),
                }
            )

    # Search issues
    for repo, issues in docs_cache[DOCUMENTATION.ISSUES].items():
        for issue in issues:
            issue_text = (
                f"{issue.get(GITHUB_DEFS.TITLE, '')} {issue.get(GITHUB_DEFS.BODY, '')}"
            )
            if query.lower() in issue_text.lower():
                results[DOCUMENTATION.ISSUES].append(
                    {
                        SEARCH_RESULT_DEFS.NAME: f"{repo}#{issue.get(GITHUB_DEFS.NUMBER)}",
                        SEARCH_RESULT_DEFS.TITLE: issue.get(GITHUB_DEFS.TITLE),
                        SEARCH_RESULT_DEFS.URL: issue.get(GITHUB_DEFS.URL),
                        SEARCH_RESULT_DEFS.SNIPPET: mcp_utils.get_snippet(
                            issue_text, query
                        ),
                    }
                )

    # Search PRs
    for repo, prs in docs_cache[DOCUMENTATION.PRS].items():
        for pr in prs:
            pr_text = f"{pr.get(GITHUB_DEFS.TITLE, '')} {pr.get(GITHUB_DEFS.BODY, '')}"
            if query.lower() in pr_text.lower():
                results[DOCUMENTATION.PRS].append(
                    {
                        SEARCH_RESULT_DEFS.NAME: f"{repo}#{pr.get(GITHUB_DEFS.NUMBER)}",
                        SEARCH_RESULT_DEFS.TITLE: pr.get(GITHUB_DEFS.TITLE),
                        SEARCH_RESULT_DEFS.URL: pr.get(GITHUB_DEFS.URL),
                        SEARCH_RESULT_DEFS.SNIPPET: mcp_utils.get_snippet(
                            pr_text, query
                        ),
                    }
                )

    # Count total results
    total_results = (
        len(results[DOCUMENTATION.README])
        + len(results[DOCUMENTATION.WIKI])
        + len(results[DOCUMENTATION.ISSUES])
        + len(results[DOCUMENTATION.PRS])
    )

    # If too many results, return error suggesting semantic search or more precise query
    if total_results > max_exact_results:
        return {
            SEARCH_RESULT_DEFS.QUERY: query,
            SEARCH_RESULT_DEFS.SEARCH_TYPE: SEARCH_TYPES.EXACT,
            "error": f"Too many results found ({total_results} matches). Exact search returned too many matches.",
            "suggestion": (
                f"Try one of the following:\n"
                f"1. Use search_type='semantic' for better relevance ranking and fewer, more relevant results\n"
                f"2. Use a more specific query (e.g., '{query} installation' or '{query} troubleshooting')\n"
                f"3. Search specific content types: filter by README, wiki, issues, or PRs"
            ),
            "result_counts": {
                DOCUMENTATION.README: len(results[DOCUMENTATION.README]),
                DOCUMENTATION.WIKI: len(results[DOCUMENTATION.WIKI]),
                DOCUMENTATION.ISSUES: len(results[DOCUMENTATION.ISSUES]),
                DOCUMENTATION.PRS: len(results[DOCUMENTATION.PRS]),
            },
        }

    return {
        SEARCH_RESULT_DEFS.QUERY: query,
        SEARCH_RESULT_DEFS.SEARCH_TYPE: SEARCH_TYPES.EXACT,
        SEARCH_RESULT_DEFS.RESULTS: results,
        SEARCH_RESULT_DEFS.TIP: "Use search_type='semantic' for natural language queries",
    }


def _get_github_headers():
    """
    Return headers for GitHub API requests, including Authorization if GITHUB_TOKEN is set.
    """
    headers = {}

    GITHUB_TOKEN = os.getenv("GITHUB_TOKEN")
    if GITHUB_TOKEN:
        print("Using token from environment variable GITHUB_TOKEN")
        headers["Authorization"] = f"token {GITHUB_TOKEN}"
    return headers


def _format_github_issue(item):
    """
    Format a GitHub issue or PR item into a standard dict.
    """
    return {
        GITHUB_DEFS.NUMBER: item[GITHUB_DEFS.NUMBER],
        GITHUB_DEFS.TITLE: item[GITHUB_DEFS.TITLE],
        GITHUB_DEFS.STATE: item[GITHUB_DEFS.STATE],
        GITHUB_DEFS.URL: item[GITHUB_DEFS.HTML_URL],
        GITHUB_DEFS.BODY: (
            (item[GITHUB_DEFS.BODY][:500] + "...")
            if item.get(GITHUB_DEFS.BODY) and len(item[GITHUB_DEFS.BODY]) > 500
            else item.get(GITHUB_DEFS.BODY)
        ),
        GITHUB_DEFS.IS_PR: GITHUB_DEFS.PULL_REQUEST in item
        or GITHUB_DEFS.MERGED_AT in item,
    }


async def _fetch_github_items(url, filter_func=None):
    """
    Fetch and format a list of GitHub issues or PRs from a given API endpoint.
    """
    async with httpx.AsyncClient(headers=_get_github_headers()) as client:
        resp = await client.get(url)
        resp.raise_for_status()
        items = []
        for item in resp.json():
            if filter_func and not filter_func(item):
                continue
            items.append(_format_github_issue(item))
        return items



from napistu.ingestion import omnipath
from napistu.ingestion.constants import OMNIPATH_ANNOTATIONS


def test_parse_omnipath_named_annotation():
    """Test the parse_omnipath_named_annotation function."""
    annotation_str = "CORUM:4478;Compleat:HC1449;PDB:4awl;PDB:6qmp;PDB:6qmq;PDB:6qms;SIGNOR:SIGNOR-C1"
    df = omnipath._parse_omnipath_named_annotation(annotation_str)

    assert df.shape == (7, 3)
    assert list(df.columns) == [
        OMNIPATH_ANNOTATIONS.NAME,
        OMNIPATH_ANNOTATIONS.ANNOTATION,
        OMNIPATH_ANNOTATIONS.ANNOTATION_STR,
    ]
    assert df.iloc[0][OMNIPATH_ANNOTATIONS.NAME] == "CORUM"
    assert df.iloc[0][OMNIPATH_ANNOTATIONS.ANNOTATION] == "4478"
    assert all(df[OMNIPATH_ANNOTATIONS.ANNOTATION_STR] == annotation_str)


def test_parse_omnipath_annotation():
    """Test the parse_omnipath_annotation function."""
    annotation_str = "11290752;11983166;12601176"
    df = omnipath._parse_omnipath_annotation(annotation_str)

    assert df.shape == (3, 2)
    assert list(df.columns) == [
        OMNIPATH_ANNOTATIONS.ANNOTATION,
        OMNIPATH_ANNOTATIONS.ANNOTATION_STR,
    ]
    assert df.iloc[0][OMNIPATH_ANNOTATIONS.ANNOTATION] == "11290752"
    assert all(df[OMNIPATH_ANNOTATIONS.ANNOTATION_STR] == annotation_str)



"""
Core MCP server implementation for Napistu.
"""

import asyncio
import logging
from typing import Any, Dict

from mcp.server import FastMCP

from napistu.mcp import codebase, documentation, execution, health, tutorials
from napistu.mcp.config import MCPServerConfig
from napistu.mcp.constants import MCP_COMPONENTS, MCP_DEFAULTS, SEARCH_TYPES
from napistu.mcp.profiles import ServerProfile, get_profile
from napistu.mcp.semantic_search import SemanticSearch

logger = logging.getLogger(__name__)


async def initialize_components(profile: ServerProfile) -> None:
    """
    Asynchronously initialize all enabled components for the MCP server.

    Parameters
    ----------
    profile : ServerProfile
        The profile whose configuration determines which components to initialize.

    Returns
    -------
    None
    """
    config = profile.get_config()

    # Define component configurations
    component_configs = [
        ("documentation", documentation, "enable_documentation"),
        ("codebase", codebase, "enable_codebase"),
        ("tutorials", tutorials, "enable_tutorials"),
        ("execution", execution, "enable_execution"),
    ]

    # Create semantic search instance
    # this supports RAG indexing of content and search using an underlying
    # sqlite vector database (chromadb)
    semantic_search = SemanticSearch()

    # Initialize all components
    initialization_results = {}

    for name, module, config_key in component_configs:
        result = await _initialize_component(
            name, module, config_key, config, semantic_search
        )
        initialization_results[name] = result

    # Initialize health components last since they monitor the other components
    logger.info("Initializing health components")
    try:
        result = await health.initialize_components()
        initialization_results["health"] = result
        if result:
            logger.info("✅ Health components initialized successfully")
        else:
            logger.warning("⚠️ Health components initialized with issues")
    except Exception as e:
        logger.error(f"❌ Health components failed to initialize: {e}")
        initialization_results["health"] = False

    # Summary of initialization
    successful = sum(1 for success in initialization_results.values() if success)
    total = len(initialization_results)
    logger.info(
        f"Component initialization complete: {successful}/{total} components successful"
    )

    if successful == 0:
        logger.error(
            "❌ All components failed to initialize - server may not function correctly"
        )
    elif successful < total:
        logger.warning(
            "⚠️ Some components failed to initialize - server running in degraded mode"
        )


def start_mcp_server(profile_name: str, server_config: MCPServerConfig) -> None:
    """
    Start MCP server - main entry point for server startup.

    Parameters
    ----------
    profile_name : str
        Server profile to use ('local', 'remote', 'full').
    server_config : MCPServerConfig
        Validated server configuration.

    Returns
    -------
    None
        This function runs indefinitely until interrupted.

    Notes
    -----
    The server uses HTTP transport (streamable-http) for all connections.
    Components are initialized asynchronously before the server starts.
    Health components are always registered and initialized last.
    """
    # Set up logging
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger("napistu")

    logger.info("Starting Napistu MCP Server")
    logger.info(f"  Profile: {profile_name}")
    logger.info(f"  Host: {server_config.host}")
    logger.info(f"  Port: {server_config.port}")
    logger.info(f"  Server Name: {server_config.server_name}")
    logger.info(f"  Transport: {MCP_DEFAULTS.TRANSPORT}")

    # Create session context for execution components
    session_context = {}
    object_registry = {}

    # Get profile with configuration
    profile: ServerProfile = get_profile(
        profile_name,
        session_context=session_context,
        object_registry=object_registry,
        server_name=server_config.server_name,
    )

    # Create server with validated configuration
    mcp = create_server(profile, server_config)

    # Initialize components first (separate async call)
    async def init_components():
        logger.info("Initializing MCP components...")
        await initialize_components(profile)
        logger.info("✅ Component initialization complete")

    # Run initialization
    asyncio.run(init_components())

    # Debug info
    logger.info(f"Server settings: {mcp.settings}")

    logger.info("🚀 Starting MCP server...")
    logger.info(
        f"Using {MCP_DEFAULTS.TRANSPORT} transport on http://{server_config.host}:{server_config.port}"
    )

    mcp.run(transport=MCP_DEFAULTS.TRANSPORT)


def create_server(profile: ServerProfile, server_config: MCPServerConfig) -> FastMCP:
    """
    Create an MCP server based on a profile configuration and server config.

    Parameters
    ----------
    profile : ServerProfile
        Server profile to use. All configuration must be set in the profile. (Valid profiles: 'execution', 'docs', 'full')
    server_config : MCPServerConfig
        Server configuration with validated host, port, and server name.

    Returns
    -------
    FastMCP
        Configured FastMCP server instance.
    """

    config = profile.get_config()

    # Create the server with validated configuration
    mcp = FastMCP(
        server_config.server_name, host=server_config.host, port=server_config.port
    )

    # Define component configurations
    component_configs = [
        (MCP_COMPONENTS.DOCUMENTATION, documentation, "enable_documentation"),
        (MCP_COMPONENTS.CODEBASE, codebase, "enable_codebase"),
        (MCP_COMPONENTS.TUTORIALS, tutorials, "enable_tutorials"),
        (MCP_COMPONENTS.EXECUTION, execution, "enable_execution"),
    ]

    # Register all components
    for name, module, config_key in component_configs:
        _register_component(
            name,
            module,
            config_key,
            config,
            mcp,
            session_context=config.get("session_context"),
            object_registry=config.get("object_registry"),
        )

    # Always register health components
    health.register_components(mcp)
    logger.info("Registered health components")

    # Register cross-component search tool
    _register_search_all_tool(mcp)

    return mcp


# private utils


def _get_semantic_search():
    """
    Get the shared semantic search instance from any component.

    This follows the same pattern as health component's _check_semantic_search_health().
    Since components store the shared semantic search instance in their state during
    initialization, we can retrieve it from any initialized component.

    Returns
    -------
    SemanticSearch or None
        Shared semantic search instance if available, None otherwise
    """
    # Try components in order until we find one with semantic search
    component_modules = [documentation, codebase, tutorials]

    for module in component_modules:
        try:
            component = module.get_component()
            if (
                hasattr(component.state, "semantic_search")
                and component.state.semantic_search
            ):
                return component.state.semantic_search
        except Exception:
            continue

    return None


async def _initialize_component(
    name: str,
    module,
    config_key: str,
    config: dict,
    semantic_search: SemanticSearch = None,
) -> bool:
    """
    Initialize a single component with error handling.

    Parameters
    ----------
    name : str
        Component name for logging
    module : module
        Component module with get_component() function
    config_key : str
        Configuration key to check if component is enabled
    config : dict
        Server configuration
    semantic_search : SemanticSearch, optional
        Shared semantic search instance for AI-powered search capabilities.
        If None, component will operate with exact text search only.

    Returns
    -------
    bool
        True if initialization successful
    """
    if not config.get(config_key, False):
        return True  # Skip disabled components

    logger.info(f"Initializing {name} components")
    try:
        component = module.get_component()
        result = await component.safe_initialize(semantic_search)
        return result
    except Exception as e:
        logger.error(f"❌ {name.title()} components failed to initialize: {e}")
        return False


def _register_component(
    name: str, module, config_key: str, config: dict, mcp: FastMCP, **kwargs
) -> None:
    """
    Register a single component with the MCP server.

    Parameters
    ----------
    name : str
        Component name for logging
    module : module
        Component module with get_component() function or create_component() for execution
    config_key : str
        Configuration key to check if component is enabled
    config : dict
        Server configuration
    mcp : FastMCP
        FastMCP server instance
    **kwargs : dict
        Additional arguments for component creation (used by execution component)
    """
    if not config.get(config_key, False):
        return  # Skip disabled components

    logger.info(f"Registering {name} components")

    if name == MCP_COMPONENTS.EXECUTION:
        # Special handling for execution component which needs session context
        component = module.create_component(
            session_context=kwargs.get("session_context"),
            object_registry=kwargs.get("object_registry"),
        )
    else:
        component = module.get_component()

    component.register(mcp)


def _register_search_all_tool(mcp: FastMCP) -> None:
    """
    Register the cross-component search tool with the MCP server.

    Parameters
    ----------
    mcp : FastMCP
        FastMCP server instance to register the tool with
    """

    @mcp.tool()
    async def search_all(
        query: str, search_type: str = SEARCH_TYPES.SEMANTIC, n_results: int = 10
    ) -> Dict[str, Any]:
        """
        Search across all Napistu components (documentation, codebase, tutorials) with intelligent search strategy.

        Provides unified search capabilities for finding relevant content across all Napistu components
        using either AI-powered semantic search for natural language queries or exact text matching
        for precise keyword searches. Searches documentation, codebase, and tutorials simultaneously.

        **USE THIS WHEN:**
        - Looking for information that might be in documentation, codebase, or tutorials
        - Wanting comprehensive search results across all Napistu content
        - Not sure which component contains the information you need
        - Searching for Napistu concepts, features, or implementation details

        **DO NOT USE FOR:**
        - General bioinformatics concepts not specific to Napistu
        - Documentation for other tools, libraries, or frameworks
        - Academic literature or research papers
        - Questions about biological concepts that don't involve Napistu usage

        **EXAMPLE APPROPRIATE QUERIES:**
        - "how to create consensus networks"
        - "SBML file processing"
        - "pathway integration"
        - "data source ingestion"
        - "network creation functions"

        Parameters
        ----------
        query : str
            Search term or natural language question about Napistu.
            Should be specific to Napistu workflows, features, or implementation.
        search_type : str, optional
            Search strategy to use:
            - "semantic" (default): AI-powered search using embeddings
            - "exact": Traditional text matching search
            Default is "semantic".
        n_results : int, optional
            Maximum number of results to return overall (not per component).
            Results are ranked by similarity score across all components.
            Default is 10.

        Returns
        -------
        Dict[str, Any]
            Search results dictionary containing:
            - query : str
                Original search query
            - search_type : str
                Actual search type used ("semantic" or "exact")
            - results : List[Dict]
                Search results ordered by relevance. Each result contains:
                - content: The matched text content
                - component: Component name ("documentation", "codebase", or "tutorials")
                - source: Human-readable source description
                - similarity_score: Float between 0 and 1 (for semantic search)
                - metadata: Dictionary with type, name, and other metadata
            - tip : str
                Helpful guidance for improving search results

        Examples
        --------
        Natural language semantic search across all components:

        >>> results = await search_all("how to create consensus networks")
        >>> print(results["search_type"])  # "semantic"
        >>> for result in results["results"]:
        ...     component = result['component']
        ...     score = result['similarity_score']
        ...     print(f"[{component}] Score: {score:.3f} - {result['source']}")

        Exact keyword search:

        >>> results = await search_all("SBML_dfs", search_type="exact")
        >>> print(f"Found {len(results['results'])} results across components")

        Notes
        -----
        **SEARCH TYPE GUIDANCE:**
        - Use semantic (default) for conceptual queries and natural language questions
        - Use exact for precise function names, class names, or known keywords

        **RESULT INTERPRETATION:**
        - Semantic results include similarity scores (0.8-1.0 = very relevant)
        - Results are ordered by relevance across all components
        - Component labels help identify the source of each result
        - Multiple results from different components may appear for comprehensive coverage

        The function automatically handles semantic search failures by falling back
        to exact search, ensuring reliable results even if AI components are unavailable.
        """
        semantic_search = _get_semantic_search()

        if search_type == SEARCH_TYPES.SEMANTIC and semantic_search:
            # Use unified semantic search - returns top K results overall, ranked by similarity
            results = semantic_search.search_unified(query, n_results=n_results)
            return {
                "query": query,
                "search_type": SEARCH_TYPES.SEMANTIC,
                "results": results,
                "tip": "Try different phrasings if results aren't relevant, or use search_type='exact' for precise keyword matching",
            }
        else:
            # Fall back to exact search across components
            # This is a simplified fallback - could be enhanced to search all components
            all_results = []

            # Try to get components and search them individually
            component_modules = [
                (MCP_COMPONENTS.DOCUMENTATION, documentation),
                (MCP_COMPONENTS.CODEBASE, codebase),
                (MCP_COMPONENTS.TUTORIALS, tutorials),
            ]

            for component_name, module in component_modules:
                try:
                    component = module.get_component()
                    # Call the component's exact search if available
                    # This is a fallback, so we'll do a simple text search
                    if hasattr(component, "state"):
                        # For now, return a message suggesting component-specific search
                        # In a full implementation, we'd search each component's cache
                        pass
                except Exception:
                    continue

            if not all_results:
                return {
                    "query": query,
                    "search_type": SEARCH_TYPES.EXACT,
                    "results": [],
                    "tip": "Exact search across components not fully implemented. Use search_type='semantic' for cross-component search, or use component-specific search tools (search_documentation, search_codebase, search_tutorials) for exact matching.",
                }

            return {
                "query": query,
                "search_type": SEARCH_TYPES.EXACT,
                "results": all_results,
                "tip": "Use search_type='semantic' for natural language queries across all components",
            }



"""
Documentation components for the Napistu MCP server.
"""

import logging
from typing import Any, Dict

from fastmcp import FastMCP

from napistu.mcp import documentation_utils
from napistu.mcp.component_base import ComponentState, MCPComponent
from napistu.mcp.constants import (
    DOCUMENTATION,
    DOCUMENTATION_SUMMARY_DEFS,
    GITHUB_DEFS,
    HEALTH_SUMMARIES,
    MCP_COMPONENTS,
    READMES,
    REPOS_WITH_ISSUES,
    SEARCH_RESULT_DEFS,
    SEARCH_TYPES,
    WIKI_PAGES,
)
from napistu.mcp.semantic_search import SemanticSearch

logger = logging.getLogger(__name__)


class DocumentationState(ComponentState):
    """
    State management for documentation component with semantic search capabilities.

    Manages cached documentation content from multiple sources and tracks semantic
    search initialization status. Extends ComponentState to provide standardized
    health monitoring and status reporting.

    Attributes
    ----------
    docs_cache : Dict[str, Dict[str, Any]]
        Nested dictionary containing cached documentation content organized by type:
        - readme: README files from repositories
        - wiki: Wiki pages from project documentation
        - issues: GitHub issues from project repositories
        - prs: Pull requests from project repositories
        - packagedown: Package documentation sections (if any)
    semantic_search : SemanticSearch or None
        Semantic search instance for AI-powered content search, None if not initialized

    Examples
    --------
    >>> state = DocumentationState()
    >>> state.docs_cache["readme"]["install"] = "Installation guide..."
    >>> print(state.is_healthy())  # True if any content loaded
    >>> health = state.get_health_details()
    """

    def __init__(self):
        """Initialize documentation state with empty cache and no semantic search."""
        super().__init__()
        self.docs_cache: Dict[str, Dict[str, Any]] = {
            DOCUMENTATION.README: {},
            DOCUMENTATION.WIKI: {},
            DOCUMENTATION.ISSUES: {},
            DOCUMENTATION.PRS: {},
            DOCUMENTATION.PACKAGEDOWN: {},
        }
        self.semantic_search = None

    def is_healthy(self) -> bool:
        """
        Check if component has successfully loaded documentation content.

        Returns
        -------
        bool
            True if any documentation section contains content, False otherwise

        Notes
        -----
        This method checks for the presence of any content in any documentation
        category. Semantic search availability is not required for health.
        """
        return any(bool(section) for section in self.docs_cache.values())

    def get_health_details(self) -> Dict[str, Any]:
        """
        Get detailed health information including content counts and semantic search status.

        Returns
        -------
        Dict[str, Any]
            Dictionary containing:
            - readme_count : int
                Number of README files loaded
            - wiki_pages : int
                Number of wiki pages loaded
            - issues_repos : int
                Number of repositories with issues loaded
            - prs_repos : int
                Number of repositories with pull requests loaded
            - total_sections : int
                Total number of content items across all categories

        Examples
        --------
        >>> state = DocumentationState()
        >>> # ... load content ...
        >>> details = state.get_health_details()
        >>> print(f"Total content items: {details['total_sections']}")
        """
        base_details = {
            HEALTH_SUMMARIES.README_COUNT: len(self.docs_cache[DOCUMENTATION.README]),
            HEALTH_SUMMARIES.WIKI_PAGES: len(self.docs_cache[DOCUMENTATION.WIKI]),
            HEALTH_SUMMARIES.ISSUES_REPOS: len(self.docs_cache[DOCUMENTATION.ISSUES]),
            HEALTH_SUMMARIES.PRS_REPOS: len(self.docs_cache[DOCUMENTATION.PRS]),
            HEALTH_SUMMARIES.TOTAL_SECTIONS: sum(
                len(section) for section in self.docs_cache.values()
            ),
        }

        return base_details


class DocumentationComponent(MCPComponent):
    """
    MCP component for documentation management and search with semantic capabilities.

    Provides access to Napistu project documentation including README files, wiki pages,
    GitHub issues, and pull requests. Supports both exact text matching and AI-powered
    semantic search for flexible content discovery.

    The component loads documentation from multiple sources:
    - README files from GitHub repositories (raw URLs)
    - Wiki pages from project wikis
    - GitHub issues and pull requests via GitHub API
    - Optional package documentation sections

    After loading content, the component initializes semantic search capabilities
    using ChromaDB and sentence transformers for natural language queries.

    Examples
    --------
    Basic component usage:

    >>> component = DocumentationComponent()
    >>> success = await component.safe_initialize()
    >>> if success:
    ...     state = component.get_state()
    ...     print(f"Loaded {state.get_health_details()['total_sections']} items")

    Notes
    -----
    The component gracefully handles failures in individual documentation sources
    and semantic search initialization. If semantic search fails, the component
    continues to function with exact text search only.
    """

    def _create_state(self) -> DocumentationState:
        """
        Create documentation-specific state instance.

        Returns
        -------
        DocumentationState
            New state instance for managing documentation content and semantic search
        """
        return DocumentationState()

    async def initialize(self, semantic_search: SemanticSearch = None) -> bool:
        """
        Initialize documentation component with content loading and semantic indexing.

        Performs the following operations:
        1. Loads README files from configured repository URLs
        2. Fetches wiki pages from project wikis
        3. Retrieves GitHub issues and pull requests via API
        4. Initializes semantic search and indexes loaded content

        Returns
        -------
        bool
            True if at least some documentation was loaded successfully, False if
            all loading operations failed
        semantic_search : SemanticSearch
            Semantic search instance for AI-powered content search, None if not initialized

        Notes
        -----
        Individual source failures are logged as warnings but don't fail the entire
        initialization. Semantic search initialization failure is logged but doesn't
        affect the return value - the component can function without semantic search.

        The method tracks success/failure rates and provides detailed logging for
        debugging content loading issues.
        """
        success_count = 0
        total_operations = 0

        # Load README files
        logger.info("Loading README files...")
        for name, url in READMES.items():
            total_operations += 1
            try:
                content = await documentation_utils.load_readme_content(url)
                self.state.docs_cache[DOCUMENTATION.README][name] = content
                success_count += 1
                logger.debug(f"Loaded README: {name}")
            except Exception as e:
                logger.warning(f"Failed to load README {name}: {e}")

        # Load wiki pages
        logger.info("Loading wiki pages...")
        for page in WIKI_PAGES:
            total_operations += 1
            try:
                content = await documentation_utils.fetch_wiki_page(page)
                self.state.docs_cache[DOCUMENTATION.WIKI][page] = content
                success_count += 1
                logger.debug(f"Loaded wiki page: {page}")
            except Exception as e:
                logger.warning(f"Failed to load wiki page {page}: {e}")

        # Load issues and PRs
        logger.info("Loading issues and pull requests...")
        for repo in REPOS_WITH_ISSUES:
            total_operations += 2  # Issues and PRs
            try:
                issues = await documentation_utils.list_issues(repo)
                self.state.docs_cache[DOCUMENTATION.ISSUES][repo] = issues
                success_count += 1
                logger.debug(f"Loaded issues for repo: {repo}")
            except Exception as e:
                logger.warning(f"Failed to load issues for {repo}: {e}")

            try:
                prs = await documentation_utils.list_pull_requests(repo)
                self.state.docs_cache[DOCUMENTATION.PRS][repo] = prs
                success_count += 1
                logger.debug(f"Loaded PRs for repo: {repo}")
            except Exception as e:
                logger.warning(f"Failed to load PRs for {repo}: {e}")

        logger.info(
            f"Documentation loading complete: {success_count}/{total_operations} operations successful"
        )

        # Initialize semantic search if content was loaded
        content_loaded = success_count > 0
        if semantic_search and content_loaded:
            self.state.semantic_search = semantic_search  # Reference, not creation
            semantic_success = await self._initialize_semantic_search()
            logger.info(
                f"Semantic search initialization: {'✅ Success' if semantic_success else '⚠️ Failed'}"
            )

        return content_loaded

    async def _initialize_semantic_search(self) -> bool:
        """
        Index documentation content into the shared semantic search instance.

        Uses the shared semantic search instance (stored in self.state.semantic_search)
        to index this component's content into the appropriate collection.

        Returns
        -------
        bool
            True if content was successfully indexed, False if indexing failed

        Notes
        -----
        Assumes self.state.semantic_search has already been set to a valid
        SemanticSearch instance during initialize().

        Failure to index content is not considered a critical error.
        The component continues to function with exact text search if semantic
        search indexing fails.
        """
        try:
            if not self.state.semantic_search:
                logger.warning("No semantic search instance available")
                return False

            logger.info("Indexing documentation content for semantic search...")

            # Index content using the shared instance stored in component state
            self.state.semantic_search.index_content(
                MCP_COMPONENTS.DOCUMENTATION, self.state.docs_cache
            )

            logger.info("✅ Documentation content indexed successfully")
            return True

        except Exception as e:
            logger.error(f"❌ Failed to index documentation content: {e}")
            return False

    def register(self, mcp: FastMCP) -> None:
        """
        Register documentation resources and tools with the MCP server.

        Registers the following MCP endpoints:
        - Resources for accessing documentation summaries and specific content
        - Tools for searching documentation with semantic and exact modes

        Parameters
        ----------
        mcp : FastMCP
            FastMCP server instance to register endpoints with

        Notes
        -----
        The search tool automatically selects semantic search when available,
        falling back to exact search if semantic search is not initialized.
        """

        # Register existing resources (unchanged)
        @mcp.resource("napistu://documentation/summary")
        async def get_documentation_summary():
            """
            Get a comprehensive summary of all available Napistu project documentation.

            **USE THIS WHEN:**
            - Getting an overview of available Napistu documentation before searching
            - Understanding what types of Napistu content are available (READMEs, wikis, issues)
            - Checking if semantic search is available for documentation

            **DO NOT USE FOR:**
            - General bioinformatics documentation (only covers Napistu project)
            - Questions not related to Napistu implementation or usage
            - Academic literature or external tool documentation

            Returns
            -------
            Dict[str, Any]
                Dictionary containing:
                - readme_files : List[str]
                    Names of loaded README files from Napistu repositories
                - issues : List[str]
                    Repository names with loaded GitHub issues
                - prs : List[str]
                    Repository names with loaded pull requests
                - wiki_pages : List[str]
                    Names of loaded Napistu wiki pages
                - packagedown_sections : List[str]
                    Names of package documentation sections
                - semantic_search : Dict[str, bool]
                    Status of semantic search availability and indexing

            Examples
            --------
            Use this to understand what Napistu documentation is available before
            searching for specific implementation details or troubleshooting information.
            """

            summary = {
                DOCUMENTATION_SUMMARY_DEFS.README_FILES: list(
                    self.state.docs_cache[DOCUMENTATION.README].keys()
                ),
                DOCUMENTATION_SUMMARY_DEFS.ISSUES: list(
                    self.state.docs_cache[DOCUMENTATION.ISSUES].keys()
                ),
                DOCUMENTATION_SUMMARY_DEFS.PRS: list(
                    self.state.docs_cache[DOCUMENTATION.PRS].keys()
                ),
                DOCUMENTATION_SUMMARY_DEFS.WIKI_PAGES: list(
                    self.state.docs_cache[DOCUMENTATION.WIKI].keys()
                ),
                DOCUMENTATION_SUMMARY_DEFS.PACKAGEDOWN_SECTIONS: list(
                    self.state.docs_cache[DOCUMENTATION.PACKAGEDOWN].keys()
                ),
            }

            return summary

        @mcp.resource("napistu://documentation/readme/{file_name}")
        async def get_readme_content(file_name: str):
            """Get the content of a specific README file."""
            if file_name not in self.state.docs_cache[DOCUMENTATION.README]:
                return {"error": f"README file {file_name} not found"}

            return {
                "content": self.state.docs_cache[DOCUMENTATION.README][file_name],
                "format": "markdown",
            }

        @mcp.resource("napistu://documentation/issues/{repo}")
        async def get_issues(repo: str):
            """Get the list of issues for a given repository."""
            return self.state.docs_cache[DOCUMENTATION.ISSUES].get(repo, [])

        @mcp.resource("napistu://documentation/prs/{repo}")
        async def get_prs(repo: str):
            """Get the list of pull requests for a given repository."""
            return self.state.docs_cache[DOCUMENTATION.PRS].get(repo, [])

        @mcp.resource("napistu://documentation/issue/{repo}/{number}")
        async def get_issue_resource(repo: str, number: int):
            """Get a single issue by number for a given repository."""
            # Try cache first
            cached = next(
                (
                    i
                    for i in self.state.docs_cache[DOCUMENTATION.ISSUES].get(repo, [])
                    if i[GITHUB_DEFS.NUMBER] == number
                ),
                None,
            )
            if cached:
                return cached
            # Fallback to live fetch
            return await documentation_utils.get_issue(repo, number)

        @mcp.resource("napistu://documentation/pr/{repo}/{number}")
        async def get_pr_resource(repo: str, number: int):
            """Get a single pull request by number for a given repository."""
            # Try cache first
            cached = next(
                (
                    pr
                    for pr in self.state.docs_cache[DOCUMENTATION.PRS].get(repo, [])
                    if pr[GITHUB_DEFS.NUMBER] == number
                ),
                None,
            )
            if cached:
                return cached
            # Fallback to live fetch
            return await documentation_utils.get_issue(repo, number)

        # Register tools
        @mcp.tool()
        async def search_documentation(
            query: str,
            search_type: str = SEARCH_TYPES.SEMANTIC,
            n_results: int = 5,
            max_exact_results: int = 20,
        ):
            """
            Search all Napistu project documentation with intelligent search strategy.

            Provides flexible search capabilities for finding relevant Napistu documentation
            using either AI-powered semantic search for natural language queries or exact text
            matching for precise keyword searches. Covers README files, wiki pages, GitHub
            issues, and pull requests from the Napistu project.

            **USE THIS WHEN:**
            - Looking for Napistu project information, setup instructions, or usage documentation
            - Finding README content, wiki pages, GitHub issues, or pull requests
            - Researching Napistu-specific concepts, workflows, troubleshooting, or features
            - Understanding Napistu installation, configuration, or implementation details

            **DO NOT USE FOR:**
            - General bioinformatics concepts not specific to Napistu
            - Documentation for other tools, libraries, or frameworks
            - Academic literature or research papers
            - Programming concepts unrelated to Napistu implementation
            - Questions about biological concepts that don't involve Napistu usage

            **EXAMPLE APPROPRIATE QUERIES:**
            - "how to install Napistu"
            - "SBML file processing with Napistu"
            - "consensus network creation"
            - "troubleshooting pathway integration"
            - "GitHub issues about data ingestion"

            **EXAMPLE INAPPROPRIATE QUERIES:**
            - "what is systems biology" (too general, not Napistu-specific)
            - "how to use pandas" (not related to Napistu)
            - "latest research in pathway analysis" (academic, not implementation)
            - "machine learning algorithms" (unless specifically about Napistu ML features)

            Parameters
            ----------
            query : str
                Search term or natural language question about Napistu. Should be
                specific to Napistu project documentation, features, or implementation.
            search_type : str, optional
                Search strategy to use:
                - "semantic" (default): AI-powered search using embeddings
                - "exact": Traditional text matching search
                Default is "semantic".
            n_results : int, optional
                Maximum number of results to return. Results are ranked by similarity score.
                Default is 5.
            max_exact_results : int, optional
                Only applicable when search_type is "exact". If more than max_exact_results are found,
                an error will be returned rather than returning all results.

            Returns
            -------
            Dict[str, Any]
                Search results dictionary containing:
                - query : str
                    Original search query
                - search_type : str
                    Actual search type used ("semantic" or "exact")
                - results : List[Dict] or Dict[str, List]
                    Search results. Format depends on search type:
                    - Semantic: List of result dictionaries with content, metadata, source, similarity_score
                    - Exact: Dictionary organized by content type (readme, wiki, issues, prs)
                - tip : str
                    Helpful guidance for improving search results

            Examples
            --------
            Natural language semantic search for Napistu concepts:

            >>> results = await search_documentation("how to install Napistu")
            >>> print(results["search_type"])  # "semantic"
            >>> for result in results["results"]:
            ...     score = result['similarity_score']
            ...     print(f"Score: {score:.3f} - Found in: {result['source']}")

            Exact keyword search for specific Napistu terms:

            >>> results = await search_documentation("installation", search_type="exact")
            >>> print(len(results["results"]["readme"]))  # Number of matching READMEs

            Notes
            -----
            **CONTENT SCOPE:**
            This tool searches only Napistu project documentation including:
            - Official README files from Napistu repositories
            - Napistu project wiki pages with implementation details
            - GitHub issues and pull requests for troubleshooting and feature discussions
            - Package documentation specific to Napistu functionality

            **SEARCH TYPE GUIDANCE:**
            - Use semantic (default) for conceptual queries and natural language questions
            - Use exact for precise Napistu function names, error messages, or known keywords

            **RESULT INTERPRETATION:**
            - Semantic results include similarity scores (0.8-1.0 = very relevant)
            - Results may include chunked sections from long documents for precision
            - Multiple related sections may appear for comprehensive coverage

            The function automatically handles semantic search failures by falling back
            to exact search, ensuring reliable results even if AI components are unavailable.
            """
            if search_type == SEARCH_TYPES.SEMANTIC and self.state.semantic_search:
                # Use semantic search
                results = self.state.semantic_search.search(
                    query, MCP_COMPONENTS.DOCUMENTATION, n_results=n_results
                )
                return {
                    SEARCH_RESULT_DEFS.QUERY: query,
                    SEARCH_RESULT_DEFS.SEARCH_TYPE: SEARCH_TYPES.SEMANTIC,
                    SEARCH_RESULT_DEFS.RESULTS: results,
                    SEARCH_RESULT_DEFS.TIP: "Try different phrasings if results aren't relevant, or use search_type='exact' for precise keyword matching",
                }
            else:
                # Fall back to exact search
                return documentation_utils._exact_search_documentation(
                    query, self.state.docs_cache, max_exact_results
                )


# Module-level component instance
_component = DocumentationComponent()


def get_component() -> DocumentationComponent:
    """
    Get the documentation component instance.

    Returns
    -------
    DocumentationComponent
        Singleton documentation component instance for use across the MCP server.
        The same instance is returned on every call to ensure consistent state.

    Notes
    -----
    This function provides the standard interface for accessing the documentation
    component. The component must be initialized via safe_initialize() before use.
    """
    return _component



from __future__ import annotations

from napistu import constants


def test_sbo_constants():
    # all SBO terms in "MINI_SBO" set have a role
    assert set(constants.SBO_NAME_TO_ROLE.keys()) == set(
        constants.MINI_SBO_FROM_NAME.keys()
    )
    # all roles are valid
    assert [x in constants.VALID_SBO_ROLES for x in constants.SBO_NAME_TO_ROLE.values()]


def test_bqb_priorities():

    assert not any(constants.BQB_PRIORITIES[constants.IDENTIFIERS.BQB].duplicated())
    assert set(constants.BQB_PRIORITIES[constants.IDENTIFIERS.BQB]) == set(
        constants.VALID_BQB_TERMS
    )



from __future__ import annotations

from napistu.ingestion import obo


def test_formatting_obo_attributes():
    assert obo._format_entry_tuple("foo: bar: baz") == ("foo", "bar: baz")
    assert obo._format_entry_tuple("foo") is None


def test_formatting_go_isa_attributes():
    go_parents_test_entries = [
        ([], list()),
        (["foo ! bar"], [{"parent_id": "foo", "parent_name": "bar"}]),
        (
            ["foo ! bar", "fun ! baz"],
            [
                {"parent_id": "foo", "parent_name": "bar"},
                {"parent_id": "fun", "parent_name": "baz"},
            ],
        ),
    ]

    for val_list, expected in go_parents_test_entries:
        assert obo._isa_str_list_to_dict_list(val_list) == expected


################################################
# __main__
################################################

if __name__ == "__main__":
    test_formatting_obo_attributes()
    test_formatting_go_isa_attributes()



from __future__ import annotations

import functools
import os
import sys
import threading

import pandas as pd
import pytest
from pytest import fixture, skip

from napistu import consensus, indices
from napistu.constants import (
    BQB,
    EXPECTED_PW_INDEX_COLUMNS,
    IDENTIFIERS,
    MINI_SBO_FROM_NAME,
    SBML_DFS,
    SBOTERM_NAMES,
    SOURCE_SPEC,
)
from napistu.identifiers import Identifiers
from napistu.ingestion.sbml import SBML
from napistu.network.constants import NAPISTU_WEIGHTING_STRATEGIES
from napistu.network.net_create import process_napistu_graph
from napistu.network.precompute import precompute_distances
from napistu.sbml_dfs_core import SBML_dfs
from napistu.source import Source


@fixture
def model_source_stub():

    source_dict = {k: "test" for k in EXPECTED_PW_INDEX_COLUMNS | {SOURCE_SPEC.MODEL}}
    return Source(pd.DataFrame(source_dict, index=[0]))


@fixture
def sbml_path():
    test_path = os.path.abspath(os.path.join(__file__, os.pardir))
    sbml_path = os.path.join(test_path, "test_data", "R-HSA-1237044.sbml")

    if not os.path.isfile(sbml_path):
        raise ValueError(f"{sbml_path} not found")
    return sbml_path


@fixture
def sbml_model(sbml_path):
    sbml_model = SBML(sbml_path)
    return sbml_model


@fixture
def sbml_dfs(sbml_model, model_source_stub):
    sbml_dfs = SBML_dfs(sbml_model, model_source=model_source_stub)
    return sbml_dfs


@fixture
def sbml_dfs_w_data(sbml_dfs):
    """Create an sbml_dfs fixture with test data for reactions and species."""

    # Add test reactions data using real indices from the sbml_dfs
    reaction_indices = (
        sbml_dfs.reactions.index[:3]
        if len(sbml_dfs.reactions) >= 3
        else sbml_dfs.reactions.index
    )
    reactions_data = pd.DataFrame(
        {
            "rxn_attr_int": [0, 1, 2],
            "rxn_attr_float": [0.1, 0.2, 0.3],
            "rxn_attr_bool": [True, False, True],
            "rxn_attr_string": ["a", "b", "c"],
        },
        index=reaction_indices,
    )
    sbml_dfs.add_reactions_data("rxn_data", reactions_data)

    # Add test species data using real indices from the sbml_dfs
    species_indices = (
        sbml_dfs.species.index[:3]
        if len(sbml_dfs.species) >= 3
        else sbml_dfs.species.index
    )
    species_data = pd.DataFrame(
        {
            "spec_attr_int": [3, 4, 5],
            "spec_attr_float": [0.4, 0.5, 0.6],
            "spec_attr_bool": [False, False, True],
            "spec_attr_string": ["A", "B", "C"],
        },
        index=species_indices,
    )
    sbml_dfs.add_species_data("spec_data", species_data)

    return sbml_dfs


@fixture
def pw_index_metabolism():
    """Create a pathway index for metabolism test data."""
    test_path = os.path.abspath(os.path.join(__file__, os.pardir))
    test_data = os.path.join(test_path, "test_data")
    return indices.PWIndex(os.path.join(test_data, "pw_index_metabolism.tsv"))


@fixture
def sbml_dfs_dict_metabolism(pw_index_metabolism):
    """Create a dictionary of SBML_dfs objects from metabolism test data."""
    return consensus.construct_sbml_dfs_dict(pw_index_metabolism)


@fixture
def sbml_dfs_metabolism(sbml_dfs_dict_metabolism, pw_index_metabolism):
    """Create a consensus SBML_dfs model from metabolism test data."""
    return consensus.construct_consensus_model(
        sbml_dfs_dict_metabolism, pw_index_metabolism
    )


@fixture
def sbml_dfs_glucose_metabolism(model_source_stub):
    test_path = os.path.abspath(os.path.join(__file__, os.pardir))
    test_data = os.path.join(test_path, "test_data")
    sbml_path = os.path.join(test_data, "reactome_glucose_metabolism.sbml")

    sbml_model = SBML(sbml_path)
    sbml_dfs = SBML_dfs(sbml_model, model_source_stub)

    return sbml_dfs


@pytest.fixture
def minimal_valid_sbml_dfs(model_source_stub):
    """Create a minimal valid SBML_dfs object for testing."""
    blank_id = Identifiers([])
    source = Source.empty()

    sbml_dict = {
        SBML_DFS.COMPARTMENTS: pd.DataFrame(
            {
                SBML_DFS.C_NAME: ["cytosol"],
                SBML_DFS.C_IDENTIFIERS: [blank_id],
                SBML_DFS.C_SOURCE: [source],
            },
            index=pd.Index(["C00001"], name=SBML_DFS.C_ID),
        ),
        SBML_DFS.SPECIES: pd.DataFrame(
            {
                SBML_DFS.S_NAME: ["ATP"],
                SBML_DFS.S_IDENTIFIERS: [blank_id],
                SBML_DFS.S_SOURCE: [source],
            },
            index=pd.Index(["S00001"], name=SBML_DFS.S_ID),
        ),
        SBML_DFS.COMPARTMENTALIZED_SPECIES: pd.DataFrame(
            {
                SBML_DFS.SC_NAME: ["ATP [cytosol]"],
                SBML_DFS.S_ID: ["S00001"],
                SBML_DFS.C_ID: ["C00001"],
                SBML_DFS.SC_SOURCE: [source],
            },
            index=pd.Index(["SC00001"], name=SBML_DFS.SC_ID),
        ),
        SBML_DFS.REACTIONS: pd.DataFrame(
            {
                SBML_DFS.R_NAME: ["test_reaction"],
                SBML_DFS.R_IDENTIFIERS: [blank_id],
                SBML_DFS.R_SOURCE: [source],
                SBML_DFS.R_ISREVERSIBLE: [False],
            },
            index=pd.Index(["R00001"], name=SBML_DFS.R_ID),
        ),
        SBML_DFS.REACTION_SPECIES: pd.DataFrame(
            {
                SBML_DFS.R_ID: ["R00001"],
                SBML_DFS.SC_ID: ["SC00001"],
                SBML_DFS.STOICHIOMETRY: [1.0],
                SBML_DFS.SBO_TERM: ["SBO:0000011"],
            },
            index=pd.Index(["RSC00001"], name=SBML_DFS.RSC_ID),
        ),
    }

    return SBML_dfs(sbml_dict, model_source_stub)


@fixture
def napistu_graph(sbml_dfs):
    """
    Pytest fixture to create a NapistuGraph from sbml_dfs with directed=True and topology weighting.
    """
    return process_napistu_graph(sbml_dfs, directed=True, weighting_strategy="topology")


@fixture
def napistu_graph_undirected(sbml_dfs):
    """
    Pytest fixture to create a NapistuGraph from sbml_dfs with directed=False and topology weighting.
    """
    return process_napistu_graph(
        sbml_dfs,
        directed=False,
        weighting_strategy=NAPISTU_WEIGHTING_STRATEGIES.TOPOLOGY,
    )


@fixture
def napistu_graph_metabolism(sbml_dfs_metabolism):
    """
    Pytest fixture to create a NapistuGraph from sbml_dfs_glucose_metabolism with directed=True and topology weighting.
    """
    return process_napistu_graph(
        sbml_dfs_metabolism,
        directed=True,
        weighting_strategy=NAPISTU_WEIGHTING_STRATEGIES.TOPOLOGY,
    )


@fixture
def precomputed_distances_metabolism(napistu_graph_metabolism):
    """
    Pytest fixture to create precomputed distances from the glucose metabolism napistu_graph.
    """
    return precompute_distances(
        napistu_graph_metabolism, max_steps=30000, max_score_q=1
    )


@fixture
def species_identifiers_metabolism(sbml_dfs_metabolism):
    """
    Pytest fixture to create a species identifiers DataFrame from sbml_dfs_metabolism.
    This creates a DataFrame with the structure expected by validate_assets.
    """

    return sbml_dfs_metabolism.get_characteristic_species_ids()


@pytest.fixture
def reaction_species_examples():
    """
    Pytest fixture providing a dictionary of example reaction species DataFrames for various test cases.
    """

    d = dict()
    d["valid_interactor"] = pd.DataFrame(
        {
            SBML_DFS.SBO_TERM: [
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.INTERACTOR],
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.INTERACTOR],
            ],
            SBML_DFS.SC_ID: ["sc1", "sc2"],
            SBML_DFS.STOICHIOMETRY: [0, 0],
        }
    ).set_index(SBML_DFS.SBO_TERM)

    d["invalid_interactor"] = pd.DataFrame(
        {
            SBML_DFS.SBO_TERM: [
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.INTERACTOR],
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.PRODUCT],
            ],
            SBML_DFS.SC_ID: ["sc1", "sc2"],
            SBML_DFS.STOICHIOMETRY: [0, 0],
        }
    ).set_index(SBML_DFS.SBO_TERM)

    d["sub_and_prod"] = pd.DataFrame(
        {
            SBML_DFS.SBO_TERM: [
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.REACTANT],
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.PRODUCT],
            ],
            SBML_DFS.SC_ID: ["sub", "prod"],
            SBML_DFS.STOICHIOMETRY: [-1, 1],
        }
    ).set_index(SBML_DFS.SBO_TERM)

    d["stimulator"] = pd.DataFrame(
        {
            SBML_DFS.SBO_TERM: [
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.REACTANT],
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.PRODUCT],
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.STIMULATOR],
            ],
            SBML_DFS.SC_ID: ["sub", "prod", "stim"],
            SBML_DFS.STOICHIOMETRY: [-1, 1, 0],
        }
    ).set_index(SBML_DFS.SBO_TERM)

    d["all_entities"] = pd.DataFrame(
        {
            SBML_DFS.SBO_TERM: [
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.REACTANT],
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.PRODUCT],
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.STIMULATOR],
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.CATALYST],
            ],
            SBML_DFS.SC_ID: ["sub", "prod", "stim", "cat"],
            SBML_DFS.STOICHIOMETRY: [-1, 1, 0, 0],
        }
    ).set_index(SBML_DFS.SBO_TERM)

    d["no_substrate"] = pd.DataFrame(
        {
            SBML_DFS.SBO_TERM: [
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.PRODUCT],
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.STIMULATOR],
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.STIMULATOR],
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.INHIBITOR],
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.CATALYST],
            ],
            SBML_DFS.SC_ID: ["prod", "stim1", "stim2", "inh", "cat"],
            SBML_DFS.STOICHIOMETRY: [1, 0, 0, 0, 0],
        }
    ).set_index(SBML_DFS.SBO_TERM)

    d["single_species"] = pd.DataFrame(
        {
            SBML_DFS.SBO_TERM: [MINI_SBO_FROM_NAME[SBOTERM_NAMES.PRODUCT]],
            SBML_DFS.SC_ID: ["lone_prod"],
            SBML_DFS.STOICHIOMETRY: [1],
        }
    ).set_index(SBML_DFS.SBO_TERM)

    d["activator_and_inhibitor_only"] = pd.DataFrame(
        {
            SBML_DFS.SBO_TERM: [
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.STIMULATOR],  # activator
                MINI_SBO_FROM_NAME[SBOTERM_NAMES.INHIBITOR],  # inhibitor
            ],
            SBML_DFS.SC_ID: ["act", "inh"],
            SBML_DFS.STOICHIOMETRY: [0, 0],
        }
    ).set_index(SBML_DFS.SBO_TERM)

    return d


# Define custom markers for platforms
def pytest_configure(config):
    config.addinivalue_line("markers", "skip_on_windows: mark test to skip on Windows")
    config.addinivalue_line("markers", "skip_on_macos: mark test to skip on macOS")
    config.addinivalue_line(
        "markers", "unix_only: mark test to run only on Unix/Linux systems"
    )


# Define platform conditions
is_windows = sys.platform == "win32"
is_macos = sys.platform == "darwin"
is_unix = not (is_windows or is_macos)


# Apply skipping based on platform
def pytest_runtest_setup(item):
    # Skip tests marked to be skipped on Windows
    if is_windows and any(
        mark.name == "skip_on_windows" for mark in item.iter_markers()
    ):
        skip("Test skipped on Windows")

    # Skip tests marked to be skipped on macOS
    if is_macos and any(mark.name == "skip_on_macos" for mark in item.iter_markers()):
        skip("Test skipped on macOS")

    # Skip tests that should run only on Unix
    if not is_unix and any(mark.name == "unix_only" for mark in item.iter_markers()):
        skip("Test runs only on Unix systems")


def skip_on_timeout(timeout_seconds):
    """Cross-platform decorator that skips a test if it takes longer than timeout_seconds"""

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            result = [None]
            exception = [None]
            finished = [False]

            def target():
                try:
                    result[0] = func(*args, **kwargs)
                    finished[0] = True
                except Exception as e:
                    exception[0] = e
                    finished[0] = True

            thread = threading.Thread(target=target)
            thread.daemon = True
            thread.start()
            thread.join(timeout_seconds)

            if not finished[0]:
                # Thread is still running, timeout occurred
                pytest.skip(f"Test skipped due to timeout ({timeout_seconds}s)")

            if exception[0]:
                raise exception[0]

            return result[0]

        return wrapper

    return decorator


pytest.skip_on_timeout = skip_on_timeout


@pytest.fixture
def sbml_dfs_characteristic_test_data(model_source_stub):
    """
    Create an SBML_dfs object with test data specifically designed for testing
    characteristic species functionality.
    """
    # Create mock species identifiers data with different BQB types
    # This matches the structure returned by get_identifiers method
    # Based on actual test results: characteristic_only=True gives (1,3), characteristic_only=False gives (1,5)
    mock_species_ids = pd.DataFrame(
        {
            SBML_DFS.S_ID: ["s1", "s1", "s1", "s1", "s1"],
            IDENTIFIERS.IDENTIFIER: [
                "P12345",  # uniprot - IS (characteristic)
                "CHEBI:15377",  # chebi - IS (characteristic)
                "GO:12345",  # go - HAS_PART (characteristic)
                "P67890",  # uniprot - HAS_VERSION (non-characteristic)
                "P67890",  # chebi - ENCODES (non-characteristic)
            ],
            IDENTIFIERS.ONTOLOGY: ["uniprot", "chebi", "go", "uniprot", "chebi"],
            IDENTIFIERS.BQB: [
                BQB.IS,  # characteristic
                BQB.IS,  # characteristic
                BQB.HAS_PART,  # characteristic
                BQB.HAS_VERSION,  # non-characteristic
                BQB.ENCODES,  # non-characteristic
            ],
        }
    )

    # Create minimal required tables for SBML_dfs
    compartments = pd.DataFrame(
        {SBML_DFS.C_NAME: ["cytosol"], SBML_DFS.C_IDENTIFIERS: [None]}, index=["C1"]
    )
    compartments.index.name = SBML_DFS.C_ID

    # Create proper identifiers for the species from the mock data
    identifiers_list = []
    for _, row in mock_species_ids.iterrows():
        identifiers_list.append(
            {
                "identifier": row[IDENTIFIERS.IDENTIFIER],
                "ontology": row[IDENTIFIERS.ONTOLOGY],
                "bqb": row[IDENTIFIERS.BQB],
            }
        )
    species_identifiers = Identifiers(identifiers_list)

    species = pd.DataFrame(
        {
            SBML_DFS.S_NAME: ["A"],
            SBML_DFS.S_IDENTIFIERS: [species_identifiers],
            SBML_DFS.S_SOURCE: [None],
        },
        index=["s1"],
    )
    species.index.name = SBML_DFS.S_ID

    compartmentalized_species = pd.DataFrame(
        {
            SBML_DFS.SC_NAME: ["A [cytosol]"],
            SBML_DFS.S_ID: ["s1"],
            SBML_DFS.C_ID: ["C1"],
            SBML_DFS.SC_SOURCE: [None],
        },
        index=["SC1"],
    )
    compartmentalized_species.index.name = SBML_DFS.SC_ID

    reactions = pd.DataFrame(
        {
            SBML_DFS.R_NAME: ["rxn1"],
            SBML_DFS.R_IDENTIFIERS: [None],
            SBML_DFS.R_SOURCE: [None],
            SBML_DFS.R_ISREVERSIBLE: [False],
        },
        index=["R1"],
    )
    reactions.index.name = SBML_DFS.R_ID

    reaction_species = pd.DataFrame(
        {
            SBML_DFS.R_ID: ["R1"],
            SBML_DFS.SC_ID: ["SC1"],
            SBML_DFS.STOICHIOMETRY: [1],
            SBML_DFS.SBO_TERM: ["SBO:0000459"],
        },
        index=["RSC1"],
    )
    reaction_species.index.name = SBML_DFS.RSC_ID

    sbml_dict = {
        SBML_DFS.COMPARTMENTS: compartments,
        SBML_DFS.SPECIES: species,
        SBML_DFS.COMPARTMENTALIZED_SPECIES: compartmentalized_species,
        SBML_DFS.REACTIONS: reactions,
        SBML_DFS.REACTION_SPECIES: reaction_species,
    }

    sbml_dfs = SBML_dfs(sbml_dict, model_source_stub, validate=False, resolve=False)

    # Store the mock data for use in tests
    sbml_dfs._mock_species_ids = mock_species_ids

    return sbml_dfs
